# AlignmentSegmentView: returns a pair (seq1,seq2), where seq1 is an ungapped subsequence
# from the base genome, and seq2 is the gapped aligned sequence from the second genome
#
# Parameters: none
# Parentviews: [AlignmentParView, AlignmentView, ChromosomeView (of base)]


import dbview
import Seg

import re


class AlignmentSegmentView(dbview.View):

    # NOTE: Module name and class name should be identical (see View._getInstanceCode)
    
    def createView(self, paramDict, parentViews, viewDB, id):
        super(AlignmentSegmentView,self).createView(paramDict, parentViews, viewDB, id)


    def loadView(self, paramDict, parentViews, viewDB, id):
        super(AlignmentSegmentView,self).loadView(paramDict, parentViews, viewDB, id)


    def init(self):
        super(AlignmentSegmentView,self).init()
        assert( self.parentViews[0].name == "AlignmentParView" )
        assert( self.parentViews[1].name == "AlignmentView" )
        assert( self.parentViews[2].name == "ChromosomeView" )
        # get segments
        apv = self.parentViews[0]
        startidx = apv.fields['start1']
        endidx = apv.fields['end1']
        self.segs = [ (apv[i][startidx],apv[i][endidx],i) for i in range(len(apv)) ]
        self.segs.sort()
        self.rsegs = [ (self.segs[i][1], i) for i in range(len(self.segs)) ]
        self.rsegs.sort()
        

    def __del__(self):
        pass


    def __getslice__(self,i,j):
        if i>=j:
            return ''
        # Find first seg with seg[1]>=i
        (a,b) = (0,len(self.rsegs))
        while a!=b:
            mid = (a+b)//2
            if self.rsegs[mid][0] < i:
                a = mid+1
            else:
                b = mid
        # Retrieve index of this segment within list sorted w.r.t. 0-th component
        if a<len(self.rsegs):
            a = self.rsegs[a][1]
        # Initialize
        alist = []       # alignments:  (left pos, right pos, seq1, seq2, oppchrom, oppseg)
        rlist = [(i,j)]  # remainder
        # Data to localize opposite segment
        opp = []
        oppseg = []
        oppstrand = 0
        oppchrom = ""
        segstart = segend = -99
        # Loop until seg[0]>=j
        while a<len(self.segs) and self.segs[a][0]<j:
            sidx = self.segs[a][2]
            # See if segment overlaps any of the remainder
            ridx = 0
            while ridx < len(rlist):
                r = rlist[ridx]
                intersect = Seg.intersection(self.segs[a][0:2],r)
                if not Seg.empty(intersect):
                    # It does
                    align = self.parentViews[1][sidx]
                    # Sanity check
                    seq1 = re.sub('-','',align[0])
                    try:
                        assert( len(seq1) == Seg.len(self.segs[a]) )
                        if ( seq1.upper() != self.parentViews[2][ self.segs[a][0]:self.segs[a][1] ].upper() ):
                            print "*** WARNING *** AlignmentSegmentView: Sequences do not match"
                            print seq1
                            print self.parentViews[2][ self.segs[a][0]:self.segs[a][1] ].upper()
                    except:
                        print "*** ERROR *** AlignmentSegmentView: Sequences have nonidentical lengths"                        
                        print len(seq1)
                        print Seg.len(self.segs[a])
                        print seq1
                        print self.parentViews[2][ self.segs[a][0]:self.segs[a][1] ].upper() 
                        raise
                    # Calculate intersection, on the aligned strand, ignoring any non-pairing bases
                    firstbase = -1                                      # start of opposite intersection; chromosome based
                    lastbase = -1
                    apv = self.parentViews[0]
                    beginbase = intersect[0]-self.segs[a][0]            # start of intersection; string-based
                    endbase = intersect[1]-self.segs[a][0]              # end
                    curbase = 0                                         # current position; string-based
                    curopposite = apv[sidx][ apv.fields['start2'] ]     # current opposite position; chromosome-based
                    chromopposite = apv[sidx][ apv.fields['chr2'] ]     
                    strandopposite = apv[sidx][ apv.fields['strand2'] ]
                    haveseg = 0
                    if len(align[0]) >= beginbase and 0 < endbase and beginbase < endbase:
                        for idx in range(len(align[0])):
                            if curbase >= beginbase and curbase < endbase and align[0][idx] != '-' and align[1][idx] != '-':
                                if firstbase == -1:
                                    firstbase = curopposite
                                lastbase = curopposite
                                if segend == curopposite-1:
                                    segend = curopposite
                                else:
                                    if segstart != -99:
                                        oppseg.append((segstart,segend+1))
                                    segstart = segend = curopposite
                                #opp.append( curopposite )
                                haveseg = 1
                            if align[0][idx] != '-':
                                curbase += 1
                            if align[1][idx] != '-':
                                curopposite += 1
                    # Add to results
                    if haveseg:                                                   # we have a segment
                        if oppchrom == "":                                        # no data yet
                            oppstrand = strandopposite
                            oppchrom = chromopposite
                            oppstart = curopposite
                        else:
                            if oppchrom == chromopposite and oppstrand == strandopposite:
                                pass
                            else:
                                oppchrom = "?"
                    # Remove gaps in base string
                    seq2 = ''.join([align[1][idx] for idx in range(len(align[1])) if align[0][idx]!='-'])
                    seq1 = re.sub('-','',align[0])
                    # Get substrings
                    seq1 = seq1[intersect[0]-self.segs[a][0] : intersect[1]-self.segs[a][0]]
                    seq2 = seq2[intersect[0]-self.segs[a][0] : intersect[1]-self.segs[a][0]]
                    # Append data
                    alist.append( (intersect[0], intersect[1], seq1, seq2, chromopposite, strandopposite, firstbase, lastbase) )
                    # Compute remaining segments
                    segl = Seg.minusL( r, intersect )
                    segr = Seg.minusR( r, intersect )
                    # Make new rlist
                    del rlist[ridx]
                    if not Seg.empty(segl):
                        rlist.append(segl)
                    if not Seg.empty(segr):
                        rlist.append(segr)
                    # Do not increase ridx
                else:
                    ridx += 1
            a += 1
        # Build result.  Add sentinels
        alist.append( (j,j,'','','none','+',-1,-1) )
        alist.sort()
        if alist[0][0] != i:
            alist.append( (i,i,'','','none','+',-1,-1) )
            alist.sort()
        seq1 = []
        seq2 = []
        oppcoords = []
        for i in range(len(alist)-1):
            if len(alist[i]) > 0:
                seq1.append(alist[i][2])
                seq2.append(alist[i][3])
                oppcoords.append( (alist[i][4],alist[i][5],(alist[i][6],alist[i][7]-1)) )
            seg = (alist[i][1],alist[i+1][0])
            gap = ''.join(['-' for i in range(seg[0],seg[1])])
            # get underlying missing data
            seq = self.parentViews[2][seg[0]:seg[1]]
            seq1.append(seq)
            seq2.append(gap)
        # convert list of indices into list of segments
        if segstart != -99:
            oppseg.append((segstart,segend+1))
        #oppseg = []
        #if opp != []:
        #    curstart = curend = opp[0]
        #    for p in opp[1:]:
        #        if p == curend+1:
        #            curend = p
        #        else:
        #            oppseg.append((curstart,curend+1))
        #            curstart = curend = p
        #    oppseg.append((curstart,curend+1))
        #print oppseg
        self.opposite = (oppchrom, oppstrand, oppseg)
        self.oppcoords = oppcoords
        return ( ''.join(seq1), ''.join(seq2) )


