# MAFXParView, creates view on alignments (without data) on .maf files
#
# Assumes that input .maf file is based on the 'base' species (see below).
# Builds a data structure that filters the required species from a big .maf file,
#  and concatenates multiple alignment blocks if the required species show conserved
#  synteny

# filename:        must be single filename (.maf or .xmfa file)
# specieslabels:   e.g. {'hg17':0,'mm5':1, 'rn3': 2}.   Numbers should be 0-based and consecutive
# gapthreshold:    max number of nucs allowed to be missing beween blocks
# base:            index of base species
# type:            "xmfa" or "maf", default "maf"
# Underlying view: chromosome of base species
#
# NOTE!!!  The labels for the start,end,strand etc. are numbered from 1 rather than 0!!!
# This is to be consistent with AlignmentParView which only has start1/start2 etc.

# Like MAFParView, MAFXParView *do* give the End positions (AlignmentParView returns length in the End fields)
#
# .MAF Lines look like:
# s (species).(chromosome) (start) (length) (strand, + or -) (chromosome length) (sequence and gaps)
#
# .XMFA lines look like:
# >(number identifying species):(start)-(end) (strand, + or -) (chromosome) (sequence and gaps)
#

from __future__ import generators

import dbview,filez


class MAFXParView(dbview.View):

    # NOTE: Module name and class name should be identical (see View._getInstanceCode)
    
    def createView(self, paramDict, parentViews, viewDB, id):
        super(MAFXParView,self).createView(paramDict, parentViews, viewDB, id)
        # load actual data
        self.alist = []
        filename = paramDict['filename']
        self.labels = paramDict['specieslabels']
        self.base = paramDict['base']
        gapthreshold = paramDict['gapthreshold']
        if paramDict.has_key('type'):
            self.type = paramDict['type']
        else:
            self.type = "maf"
        assert self.parentViews[0].name == "ChromosomeView"
        # get chromosome length, for sanity check
        basechromlen = len(self.parentViews[0])
        # initialise first block to empty
        curchroms = ["-nothing-" for i in self.labels]  # Non-empty sentinel; makes sure synteny gets checked
        curstartpos = [-1 for i in self.labels]         # -1 start positions makes sure this block does not get added
        curendpos = [0 for i in self.labels]
        curfplist = [[] for i in self.labels]
        curbaseposlist = []
        curfirstlast = [0,99999,0]
        oldfirstlast = curfirstlast[:]
        curglue = []
        # read .maf file; add sentinel; initialise line counter and file pointer
        handle = filez.open(filename,'r')
        lines = handle.readlines()
        lines.append("X")
        i = fp = 0
        while i < len(lines)-1:
            # Loop over lines; look for alignment ('a'); update file pointer (fp)
            line = lines[i]
            fp += len(line)
            i += 1
            if line[0] == 'a' or line[0:2] == '>1':
                # Found alignment block.  Enter inner loop interpreting each species
                if self.type == 'maf':
                    # read first data line for .maf files; it is current for .xmfa files
                    line = lines[i]
                    fp += len(line)
                    i += 1
                # Build fresh first block, in case synteny dictates to start a new block
                oldfirstlast = curfirstlast[:]
                curfirstlast[0] += 1
                newfirstlast = [0,99999,0]
                newchroms = ["" for j in self.labels]
                newstartpos = [-1 for j in self.labels]
                newendpos = [0 for j in self.labels]
                newfplist = [0 for j in self.labels]
                newbaseposlist = []
                # Start off hoping synteny is conserved
                syncons = True
                while line[0] == 's' or line[0] == '>':    # loop until separator: 'a' (maf) or '=' (xmfa)
                    # Got another species line
                    if self.type == 'maf':
                        (schar,specchrom,pos,nuclen,strand,chromlen,seq) = line.split()
                        (spec,chrom) = specchrom.split('.')
                        curfp = fp-len(line)   # points to beginning of this line
                        # new line
                        line = lines[i]
                        fp += len(line)
                        i += 1
                    else:
                        (firstbit,strand,chrom) = line.split()
                        spec = firstbit[1]
                        (pos,end) = firstbit[3:].split('-')
                        nuclen = int(end)-int(pos)
                        curfp = fp             # points to beginning of next line (start of seq data)
                        seq = ""
                        # new line
                        line = lines[i]
                        fp += len(line)
                        i += 1
                        while line[0] != '>' and line[0] != '=':
                            line = lines[i]
                            fp += len(line)
                            i += 1
                    # Check whether we're interested
                    if self.labels.has_key(spec):
                        specid = self.labels[spec]
                        # Sanity check, and update first and last block with some non-base alignment
                        if specid == self.base:
                            # Remember start and end position of base sequence for this section
                            newbaseposlist = [ (int(pos), int(pos)+int(nuclen)) ]
                            if self.type == 'maf':
                                if int(chromlen) != basechromlen:
                                    print "MAF chrom length ",chromlen," whereas base chrom length ",basechromlen
                                    assert False
                        else:
                            # We have some non-base alignment
                            newfirstlast[1] = min(newfirstlast[0], newfirstlast[1])
                            newfirstlast[2] = max(newfirstlast[0], newfirstlast[2])
                            curfirstlast[1] = min(curfirstlast[0], curfirstlast[1])
                            curfirstlast[2] = max(curfirstlast[0], curfirstlast[2])
                        # Initialize new-block variables
                        newchroms[specid] = chrom + strand
                        newstartpos[specid] = int(pos)
                        newendpos[specid] = int(pos)+int(nuclen)
                        newfplist[specid] = [curfp]
                        # Did we see this species before?
                        if curchroms[specid] != "":
                            # Check that synteny is conserved
                            if curchroms[specid] != chrom + strand:
                                syncons = False
                            if curendpos[specid] > int(pos):
                                syncons = False
                            if curendpos[specid] + gapthreshold <= int(pos):
                                syncons = False
                    else:
                        pass
                # Complete the new-block variables
                for j in range(len(self.labels)):
                    if newfplist[j] == 0:
                        newfplist[j] = [-1]
                # Decide whether to concatenate, or to publish and renew
                if syncons:
                    # Concatenate.  First get missing base sequence, if any.
                    # Assumes the base sequence lives on the positive strand.
                    seqglue = self.parentViews[0][ curendpos[ self.base ]: newstartpos[ self.base ] ]
                    curglue.append( seqglue )
                    # Update current block
                    for j in range(len(self.labels)):
                        if newendpos[j] != 0:
                            curendpos[j] = newendpos[j]
                        if curchroms[j] == "":
                            curchroms[j] = newchroms[j]
                        if curstartpos[j] == -1:
                            curstartpos[j] = newstartpos[j]
                        curfplist[j].append( newfplist[j][0] )
                    curbaseposlist.append( newbaseposlist[0] )
                else:
                    # Publish.  Move data to alist -- if we have data for all
                    pub = self._publish( oldfirstlast, curstartpos, curendpos, curchroms, curfplist, curglue, curbaseposlist )
                    if pub:
                        self.alist.append( pub )
                    # Make new block current
                    curchroms = newchroms
                    curstartpos = newstartpos
                    curendpos = newendpos
                    curfplist = newfplist
                    curglue = []
                    curfirstlast = newfirstlast
                    curbaseposlist = newbaseposlist
        # Do last block
        oldfirstlast = curfirstlast
        pub = self._publish( oldfirstlast, curstartpos, curendpos, curchroms, curfplist, curglue, curbaseposlist )
        if pub:
            self.alist.append( pub )
        del(lines)
        handle.close()
        # now cache them
        self.cache(self.alist,'alist')


    def loadView(self, paramDict, parentViews, viewDB, id):
        super(MAFXParView,self).loadView(paramDict, parentViews, viewDB, id)
        self.alist = self.uncache('alist')


    def init(self):
        super(MAFXParView,self).init()
        self.makefields()


    # Helper routine
    def _publish(self, oldfirstlast, curstartpos, curendpos, curchroms, curfplist, curglue, curbaseposlist ):
        haveall = True
        curstrand = []
        first = oldfirstlast[1]
        last = oldfirstlast[2]
        for j in range(len(curstartpos)):
            if curstartpos[j] == -1:
                haveall = False
            else:
                # Separate chromosome and strand
                curstrand.append( curchroms[j][-1] )
                curchroms[j] = curchroms[j][:-1]
                # Slice required file positions
                curfplist[j] = curfplist[j][first:last+1]
        if haveall:
            # Account for slicing away blocks that do not align to interesting species
            curstartpos[ self.base ] = curbaseposlist[first][0]
            curendpos[ self.base ] = curbaseposlist[last][1]
            return ( (curglue[first:last],
                      curstartpos,
                      curendpos,
                      curchroms,
                      curstrand,
                      curfplist) )
        else:
            return False


    def makefields(self):
        self.fields = {'glue':0}
        self.labels = self.paramDict['specieslabels']
        self.base = self.paramDict['base']
        if self.paramDict.has_key('type'):
            self.type = self.paramDict['type']
        else:
            self.type = 'maf'
        curf = 1
        numlabels = len(self.labels)
        for specie in range(numlabels):
            self.fields[ 'start'+str(specie+1) ] = curf
            self.fields[ 'end'+str(specie+1) ] = curf+numlabels
            self.fields[ 'chr'+str(specie+1) ] = curf+numlabels*2
            self.fields[ 'strand'+str(specie+1) ] = curf+numlabels*3
            self.fields[ 'fplist'+str(specie+1) ] = curf+numlabels*4
            curf += 1


    def iter(self):
        i = 0
        while 1:
            try:
                yield self[i]
            except IndexError:
                raise StopIteration
            i += 1


    def __getitem__(self,key):
        (curglue,curstartpos,curendpos,curchroms,curstrand,curfplist) = self.alist[key]
        return tuple( [curglue] + curstartpos + curendpos + curchroms + curstrand + curfplist )



    def __len__(self):
        return self.alist.__len__()
