# SegmentIndexView, creates view of segments, taken from other view(s)

#
# - either 'singleSegment':(idx,idx) retrieves pairs of numbers from tuples or lists in other view(s)
# - or 'segmentList':idx             retrieves lists of segments from other view(s)
# - or (default)                     takes segments from underlying view(s)
#
# - 'condition:[(idx,value),..]  takes only those segments (through singleSegment or segmentList) for which elt[idx]==value.
#
# shoulder: pads segment shoulders
#
# Similar to SegmentView, but has getIndex(p) method, returning the index of the segment containing p (if it exists)
#
# Does not process segments; overlapping, unordered and empty segments are kept, so that any relationship
# with another view is left undisturbed.  A sorted view is computed, in order to implement getIndex efficiently.
# Also, taking inverses is not supported
#
# 19/2/05: bugfix overlap removal
#

from __future__ import generators


import dbview
import Seg



class SegmentIndexView(dbview.View):

    # NOTE: Module name and class name should be identical (see View._getInstanceCode)
    
    def createView(self, paramDict, parentViews, viewDB, id):
        super(SegmentIndexView,self).createView(paramDict, parentViews, viewDB, id)
        if paramDict.has_key('singleSegment'):
            # value: two indices, start and end of segment
            getSegs = self._getSingle(paramDict['singleSegment'])
        elif paramDict.has_key('segmentList'):
            # value: single index, list of segments
            getSegs = self._getList(paramDict['segmentList'])
        else:
            # assume that underlying view is a sequence of segments
            getSegs = self._getSingle([0,1])
        if paramDict.has_key('shoulder'):
            shoulder = paramDict['shoulder']
        else:
            shoulder = 0
        if paramDict.has_key('condition'):
            self.condition = paramDict['condition']
        else:
            self.condition = []
        #assert len(self.parentViews) > 0
        self.segList = []
        self.sortedSegList = []
        for view in self.parentViews:
            for i in range(len(view)):
                obj = view[i]
                segs = getSegs(obj)
                for seg in segs:
                    if not Seg.empty(seg):
                        self.segList.append((seg[0]-shoulder,seg[1]+shoulder))
                        if not Seg.empty(self.segList[-1]):
                            self.sortedSegList.append((self.segList[-1],i))
                    else:
                        self.segList.append(Seg.empty)
        self.sortedSegList.sort()
        # store for indexIntersection
        self.overlappingSortedSegList = self.sortedSegList[:]
        # now remove overlaps
        if len(self.sortedSegList) > 0:
            curseg = self.sortedSegList[0][0]
        for i in range(1,len(self.sortedSegList)):
            if Seg.overlap(curseg,self.sortedSegList[i][0]):
                self.sortedSegList[i] = (Seg.minusR( self.sortedSegList[i][0],curseg ), self.sortedSegList[i][1] )
                if not Seg.empty( self.sortedSegList[i][0] ):
                    curseg = (curseg[0],self.sortedSegList[i][0][1])
            else:
                curseg = self.sortedSegList[i][0]
        # remove empties
        self.sortedSegList = [seg for seg in self.sortedSegList if not Seg.empty(seg[0])]
        self.cache(self.segList,'seglist')
        self.cache(self.sortedSegList,'sortedseglist')
        self.cache(self.overlappingSortedSegList,'overlapping')


    def loadView(self, paramDict, parentViews, viewDB, id):
        super(SegmentIndexView,self).loadView(paramDict, parentViews, viewDB, id)
        self.segList = self.uncache('seglist')
        self.sortedSegList = self.uncache('sortedseglist')
        self.overlappingSortedSegList = self.uncache('overlapping')


    def init(self):
        super(SegmentIndexView,self).init()



    def _getOverlapCeilingIndex(self,pt):
        # returns index to first segment starting beyond pt
        lo = 0
        hi = len(self.overlappingSortedSegList)
        # Invariant: lo <= [pt] <= hi
        # [pt] is leftmost segment strictly above pt
        while lo < hi:
            mid = (lo+hi)//2
            # lo <= mid < hi
            if self.overlappingSortedSegList[mid][0][0] >= pt:
                # strictly above pt
                hi = mid
            else:
                lo = mid+1
        return lo



    def _getFloorIndex(self,pt):
        # returns index to segment containing pt, or first index of segment beyond pt if there's no such elt
        lo = 0
        hi = len(self.sortedSegList)
        # Invariant: lo <= [pt] <= hi
        while lo < hi:
            mid = (lo+hi)//2
            # lo <= mid < hi
            if self.sortedSegList[mid][0][1] <= pt:
                # pt is outside (above) segment mid
                lo = mid+1
            else:
                # pt is inside or below segment mid
                hi = mid
        return lo


    def _getCeilingIndex(self,pt):
        # returns index to segment just above that containing any pt-1,
        # or first index of segment beyond pt-1 if there's no such index
        lo = 0
        hi = len(self.sortedSegList)
        # Invariant: lo <= [pt] <= hi
        # [pt] is leftmost segment strictly above of pt
        while lo < hi:
            mid = (lo+hi)//2
            # lo <= mid < hi
            if self.sortedSegList[mid][0][0] > pt:
                # strictly above pt
                hi = mid
            else:
                lo = mid+1
        return lo



    def getIndex(self,pt):
        # returns index to segment containing pt, or -1 if there's no such elt
        # Note: Both sortedSegList[i][0][0] and sSL[i][0][1] are increasing
        lo = 0
        hi = len(self.sortedSegList)
        while lo < hi:
            mid = (lo+hi)//2
            if self.sortedSegList[mid][0][1] <= pt:
                lo = mid+1
            else:
                hi = mid
        if lo==len(self.sortedSegList):
            return -1
        if Seg.overlapelt(self.sortedSegList[lo][0],pt):
            # return index into user-visible segment list
            return self.sortedSegList[lo][1]
        return -1



    def getNearestIndex(self,pt):
        # returns index to segment containing pt, or nearest segment if there's no such elt; or -1 if no segments at all
        lo = 0
        hi = len(self.sortedSegList)
        if hi==0:
            return -1
        while lo < hi:
            mid = (lo+hi)//2
            if self.sortedSegList[mid][0][1] <= pt:
                lo = mid+1
            else:
                hi = mid
        if lo==len(self.sortedSegList):
            return self.sortedSegList[lo-1][1]
        if Seg.overlapelt(self.sortedSegList[lo][0],pt):
            # return index into user-visible segment list
            return self.sortedSegList[lo][1]
        if lo==0:
            # first segment is above pt; so return that
            return self.sortedSegList[lo][1]
        # now return nearest
        if pt-self.sortedSegList[lo-1][0][1] < self.sortedSegList[lo][0][0]-pt:
            return self.sortedSegList[lo-1][1]
        else:
            return self.sortedSegList[lo][1]



    def overlap(self,seg):
        # returns total overlap of segment list with seg
        idx = self._getFloorIndex(seg[0])
        ovl = 0
        while idx < len(self.sortedSegList) and self.sortedSegList[idx][0][0]<seg[1]:
            ovl += Seg.len( Seg.intersection( self.sortedSegList[idx][0], seg ) )
            idx += 1
        return ovl
            


    def intersection(self,seg):
        # returns intersection of segment list with seg (or list of non-intersecting segs), as a (simple) list of segments
        if type(seg)==type([]):
            # distinguish list of segments from segment as list of integers
            if (len(seg) == 2) and (type(seg[0])==int or type(seg[0])==long):
                sl = [seg]
            else:
                sl = seg
        else:
            sl = [seg]
        inter = []
        for sg in sl:
            idx = self._getFloorIndex(sg[0])
            while idx < len(self.sortedSegList) and self.sortedSegList[idx][0][0]<sg[1]:
                inter.append( Seg.intersection( self.sortedSegList[idx][0], sg ) )
                idx += 1
        return inter
            

    def indexIntersection(self,seg):
        # returns indices of intersection of segments with seg
        # Note:  This routine is heuristic.  It will fail to return very long segments that overlap
        #  seg in both directions.
        inter = []
        idx = self._getOverlapCeilingIndex(seg[1])-1
        heuristic = startheuristic = 15
        while idx >=0 and heuristic >= 0:
            if not Seg.empty( Seg.intersection( self.overlappingSortedSegList[idx][0], seg ) ):
                inter.append( self.overlappingSortedSegList[idx][1] )
                heuristic = startheuristic
            idx -= 1
            heuristic -= 1
        return inter
            

    def _getSingle(self,seg):
        # returns list with single segment
        def _getSegs(elt):
            for c in self.condition:
                if not elt[c[0]]==c[1]:
                    return []
            return [(elt[seg[0]],elt[seg[1]])]
        return _getSegs


    def _getList(self,segListIdx):
        # returns list of segments
        def _getSegs(elt):
            for c in self.condition:
                if not elt[c[0]]==c[1]:
                    return []
            return elt[segListIdx]
        return _getSegs
    

    def __iter__(self):
        return iter(self.segList)


    def __getitem__(self,key):
        return self.segList[key]


    def __len__(self):
        return len(self.segList)


