# SegmentView, creates view of segments, taken from other view(s)

#
# - either 'singleSegment':(idx,idx) retrieves pairs of numbers from tuples or lists in other view(s)
# - or 'segmentList':idx             retrieves lists of segments from other view(s)
# - or 'cPickle':file                retrieves lists of segments from cPickled data ('index': take index into sublist or dictionary)
# - or (default)                     takes segments from underlying view(s)
# - 'inverse:(left,right)        takes inverse within segment [left,right)
# - 'condition:[(idx,value),..]  takes only those segments (through singleSegment or segmentList) for which elt[idx]==value.
#

from __future__ import generators

import cPickle
import dbview
import Seg



class SegmentView(dbview.View):

    # NOTE: Module name and class name should be identical (see View._getInstanceCode)
    
    def createView(self, paramDict, parentViews, viewDB, id):
        super(SegmentView,self).createView(paramDict, parentViews, viewDB, id)
        if paramDict.has_key('singleSegment'):
            # value: two indices, start and end of segment
            getSegs = self._getSingle(paramDict['singleSegment'])
            self.listlist = self.parentViews
        elif paramDict.has_key('segmentList'):
            # value: single index, list of segments
            getSegs = self._getList(paramDict['segmentList'])
            self.listlist = self.parentViews
        elif paramDict.has_key('cPickle'):
            # value: filename
            getSegs = self._getFile(paramDict)
        else:
            # assume that underlying view is a sequence of segments
            getSegs = self._getSingle([0,1])
            self.listlist = self.parentViews
        if paramDict.has_key('shoulder'):
            shoulder = paramDict['shoulder']
        else:
            shoulder = 0
        if paramDict.has_key('condition'):
            self.condition = paramDict['condition']
        else:
            self.condition = []
        #assert len(self.listlist) > 0
        self.segList = []
        for view in self.listlist:
            for obj in view:
                if shoulder == 0:
                    self.segList.extend( getSegs( obj ) )
                else:
                    segs = getSegs(obj)
                    for seg in segs:
                        if not Seg.empty(seg):
                            self.segList.append((seg[0]-shoulder,seg[1]+shoulder))
        del self.listlist
        self.segList = [seg for seg in self.segList if not Seg.empty(seg)]
        self.segList.sort()
        # now remove overlaps, and represent list as efficiently as possible
        for i in range(1,len(self.segList)):
            if Seg.touch(self.segList[i-1],self.segList[i]):
                self.segList[i] = Seg.union( self.segList[i-1],self.segList[i] )
                self.segList[i-1] = Seg.nullSeg
        # remove empties
        self.segList = [seg for seg in self.segList if not Seg.empty(seg)]
        # now invert if asked
        self._invert()
        self.totlen = 0
        for s in self.segList:
            self.totlen += Seg.len(s)
        self.cache(self.segList,'seglist')


    def removeSegs(self, remove):
        iX = [seg for seg in remove if not Seg.empty(seg)]
        iX.sort()
        iY = []
        i = 0
        j = 0
        self.totlen = 0
        curseg = self.segList[i]
        while not Seg.empty(curseg):
            while j<len(iX) and iX[j][1] <= curseg[0]:
                j+=1
            if j<len(iX) and Seg.overlap( curseg, iX[j] ):
                new = Seg.minusL( curseg, iX[j] )
                if not Seg.empty(new):
                    iY.append(new)
                    self.totlen += Seg.len(new)
                curseg = Seg.minusR(curseg, iX[j])
                j+=1
            else:
                iY.append(curseg)
                self.totlen += Seg.len(curseg)
                curseg = Seg.nullSeg
            if Seg.empty(curseg):
                i+=1
                if i<len(self.segList):
                    curseg = self.segList[i]
        self.segList = iY
        


    def loadView(self, paramDict, parentViews, viewDB, id):
        super(SegmentView,self).loadView(paramDict, parentViews, viewDB, id)
        self.segList = self.uncache('seglist')
        self.totlen = 0
        for s in self.segList:
            self.totlen += Seg.len(s)


    def init(self):
        super(SegmentView,self).init()


    def _getSingle(self,seg):
        # returns list with single segment
        def _getSegs(elt):
            for c in self.condition:
                if not elt[c[0]]==c[1]:
                    return []
            return [(elt[seg[0]],elt[seg[1]])]
        return _getSegs


    def _getList(self,segListIdx):
        # returns list of segments
        def _getSegs(elt):
            for c in self.condition:
                if not elt[c[0]]==c[1]:
                    return []
            return elt[segListIdx]
        return _getSegs


    def _getFile(self,paramDict):
        # returns list with single segments
        file = open(paramDict['cPickle'],"r")
        segmentlist = cPickle.load(file)
        if paramDict.has_key('index'):
            segmentlist = segmentlist[paramDict['index']]
        self.listlist = [segmentlist]
        def _getSegs(elt):
            return [elt]
        return _getSegs


    def __iter__(self):
        return iter(self.segList)


    def __getitem__(self,key):
        return self.segList[key]


    def __len__(self):
        return len(self.segList)


    def totLen(self):
        return self.totlen;


