# dbview.py - a simple persistent database view system
#
# Support for cachable views on read-only
# databases, and full annotation of caches.
#
# Gerton Lunter, 31/1/2004
#

import sys
import os
import time
import random
import cPickle
import bisect

import Seg


class dbview(object):

    def __init__(self, rootfile):
        # reads global DB variables from root file, or creates file
        self.rootfile = rootfile
        self.logfile = rootfile+".log"
        self.maxcreationtime = 1.0
        random.seed()
        if not os.path.exists(self.rootfile+".present"):
            # Root file does not exist
            presentfile = open(self.rootfile+".present","w")
            presentfile.write("--present--")
            presentfile.close()
            self.rootdict = {'idCount':0,'cacheDir':'./','verbose':True,'binaryCache':True,'knownViews':[],'codes':{}}
            self._commit(self.rootfile+".first")
        else:
            # Root file does exist.  Open it
            self.rootdict = {'verbose':True}    # for _msg
            name = self._lockread(2000)
            self._unlock(name)


    def _lockread(self,maxtries = 20000):
        name = self.rootfile + "." + str( random.randrange(0,10000000) )
        # waits until dictionary is locked
        tries = 0
        done = False
        while not done and tries < maxtries:
            try:
                os.rename(self.rootfile, name)
                done = True
            except OSError:
                if self.rootdict['verbose']:
                    if tries==0:
                        print "Lockread miss... ",
                tries += 1
                if self.rootdict['verbose']:
                    print tries,
                if tries % 1000 == 0:
                    self._msg("Tried to lock root file "+str(tries)+" times, continuing...")
                time.sleep( random.random() )
        if not done:
            msg = "Tried to lock root file "+self.rootfile+" but no success, giving up."
            print msg
            self._msg(msg)
            assert False
        if tries>0:
            if self.rootdict['verbose']:
                print
            self._msg("Lockread misses: "+str(tries))
        try:
            file = open(name,'r')
            self.rootdict = cPickle.load(file)
            file.close()
        except:
            print "Error loading or unpickling "+name+".  Giving up."
            self._unlock(name)
            print "Unlocking "+name+" successful."
            raise
        return name


    def _unlock(self,name):
        os.rename(name,self.rootfile)
                

    def _commit(self,name):
        # saves global DB variables to root file.  Also relinquishes lock
        newname = name+".new"
        now = time.time()
        file = open(newname,'w')
        # Use bin=1 (or protocol=1) for efficiency, and to avoid "insecure string" errors
        cPickle.dump(self.rootdict,file,1)
        file.close()
        writetime = time.time() - now
        self._log("Root file written (time = "+str(writetime)+")")
        # Successfully written new file.  Now it's safe to remove old one
        if os.access(name,os.W_OK):
            os.unlink(name)
        else:
            if name != self.rootfile+".first":
                print "WARNING (dbview._commit): Could not unlink ",name,", proceeding."
        self._unlock(newname)



    def unsafeCommit(self):
        # If running in unsafe mode, this function is used to write the rootfile to disk
        self._commit(self.rootfile)



    def _getId(self,safe):
        # Returns new identifier.
        if safe:
            name = self._lockread()
        self.rootdict['idCount'] += 1
        if safe:
            self._commit(name)
        return self.rootdict['idCount']


    def _msg(self,msg):
        self._log(msg)
        if self.rootdict['verbose']:
            print msg


    def _log(self,msg):
        line = time.asctime()+'|'+os.environ.get('HOSTNAME')+'|'+str(os.getpid())+": "+msg
        log = open(self.logfile,"a")
        log.write(line+"\n")
        log.close()
        

    def setRootDict(self, key, value):
        if key not in self.rootdict or self.rootdict[key] != value:
            name = self._lockread()
            self.rootdict[key] = value
            self._commit(name)


    def dumpCodes(self):
        for c in self.rootdict['codes']:
            codeid = self.rootdict['codes'][c]
            views = 0
            for view in self.rootdict['knownViews']:
                if view['code']==codeid:
                    views += 1
                print "---- codeid =",codeid," numviews =",views,"----"
                print c
                print "***********************************************"
        


    def _findView(self, paramKeys, paramVals, parentIds, codeid, codeCheck):
        for view in self.rootdict['knownViews']:
            if view['paramKeys'] == paramKeys and \
                   view['paramVals'] == paramVals and \
                   view['parentIds'] == parentIds and \
                   view['id']>=0 and \
                   (view['code'] == codeid or not codeCheck):
                return view
        return None
        

    def open(self, viewClass, paramDict = {}, parentViews = [], codeCheck = True, cache = True, safe = True, globals = None):
        newView = eval( viewClass+"."+viewClass )()
        if globals:
            # dirty hack, to grant 'eval' access to global namespace; used to enable storing functions as strings rather
            # than true function pointers, which would otherwise require definitions to always be present.
            newView.setGlobals(globals)
        # Get/create identifier for code
        code = newView._getInstanceCode()
        if not code in self.rootdict['codes']:
            # Reload dictionary and check again
            if safe:
                name = self._lockread()
            if not code in self.rootdict['codes']:
                self.rootdict['codes'][code] = len(self.rootdict['codes'])
                if safe:
                    self._commit(name)
            else:
                if safe:
                    self._unlock(name)
        codeid = self.rootdict['codes'][code]
        parentIds = [view.id for view in parentViews]
        paramKeys = paramDict.keys()
        paramKeys.sort()
        paramVals = [paramDict[key] for key in paramKeys]
        # Make sure that if view depends on uncached views, it is not cached
        for pid in parentIds:
            if pid == -1:
                cache = False
        # If not caching data, simply make one with identifier -1
        if not cache:
            now = time.time()
            newView.createView(paramDict, parentViews, self, -1)
            createtime = time.time() - now
            self._log("Created uncached view for "+newView.name+" (time = "+str(createtime)+")")
            newView._flushcache(False)
            return newView
        # Check whether this view exists, and call loadView or createView as appriopriate
        view = self._findView(paramKeys, paramVals, parentIds, codeid, codeCheck)
        if view != None:
            if view.has_key('slow') and not view['slow']:
                self._msg("Rebuilding view for "+newView.name + " (id = " + str(view['id']) + ")")
                newView.createView(paramDict, parentViews, self, view['id'])
                newView._flushcache(False)
            else:
                self._msg("Loading view for " + newView.name + " (id = " + str(view['id']) + ")")
                newView.loadView(paramDict, parentViews, self, view['id'])
            return newView
        # Now create view.  First, get a safe identifier.  Reload the root dictionary in case it was changed asynchronously
        id = self._getId(safe)
        # Perhaps view was made asynchronously?
        view = self._findView(paramKeys, paramVals, parentIds, codeid, codeCheck)
        if view != None:
            if view.has_key('slow') and not view['slow']:
                self._msg("Async build: Rebuilding view for "+newView.name + " (id = " + str(view['id']) + ")")
                newView.createView(paramDict, parentViews, self, view['id'])
                newView._flushcache(False)
            else:
                self._msg("Async build: Loading view for " + newView.name + " (id = " + str(view['id']) + ")")
                newView.loadView(paramDict, parentViews, self, view['id'])
            return newView
        # Now create view
        self._msg("Creating view for " + newView.name + " (id = " + str(id) + ")")
        now = time.time()
        newView.createView(paramDict, parentViews, self, id)
        creationtime = time.time() - now
        slow = False
        if creationtime > self.maxcreationtime:
            slow = True
        self._log("View created for " + newView.name + " (time = "+str(creationtime)+" id = "+str(id)+")")
        viewData = {'id': id, 'code':codeid, 'paramKeys':paramKeys, 'paramVals':paramVals, 'parentIds':parentIds, 'slow':slow}
        # Flush cache (if it took more than 1 second to create)
        newView._flushcache(slow)
        # Next, save data in root dictionary.  Reload dictionary in case other views have been added asynchronously
        if safe:
            name = self._lockread()
        self.rootdict['knownViews'].append(viewData)
        if safe:
            self._commit(name)
        # Done!
        return newView


    def removeView(self,id):
        name = self._lockread()
        removed = 0
        for i in range(len(self.rootdict['knownViews'])):
            if self.rootdict['knownViews'][i]['id'] == id:
                self.rootdict['knownViews'][i]['id'] = -1
                self.rootdict['knownViews'][i]['code'] = -1
                removed = 1
                print "Removed id=",id
        if removed:
            self._commit(name)
        else:
            self._unlock(name)



class View(object):

    def __init__(self):
        self.name = self.__class__.__name__
        self.cachedata = []


    def _getInstanceCode(self):
        # Returns code for class.  Relies on module having
        # same name as class.
        for dir in sys.path:
            if dir == "":
                filename = self.name + ".py"
            else:
                filename = dir + os.sep + self.name + ".py"
            if os.path.exists(filename):
                file = open(filename,'r')
                code = file.read()
                file.close()
                return code
        return "<Python code not found>"


    def _invert(self):
        # Subroutine, used more often so put here
        if self.paramDict.has_key('inverse'):
            (left,right) = self.paramDict['inverse']
            self.segList.append([right,right])
            newList = []
            for i in range(len(self.segList)):
                newList.append((left,min(self.segList[i][0],right)))
                left = max(left,self.segList[i][1])
            self.segList = newList
            # remove empties
            self.segList = [seg for seg in self.segList if not Seg.empty(seg)]


    def createView(self, paramDict, parentViews, viewDB, id):
        # Creates cache.  This base class implementation just copies the initialisation variables
        self.paramDict = paramDict
        self.parentViews = parentViews
        self.viewDB = viewDB
        self.id = id
        self.init()


    def intersection(self, seg):
        # returns list of segments in underlying segment-view that overlap seg
        # part of code taken from bisect.py
        hi = len(self)
        lo = 0
        while lo < hi:
            mid = (lo+hi)//2
            if self[mid][1] < seg[0]:
                lo = mid+1
            else:
                hi = mid
        # lo is first elt to be considered
        first = lo
        hi = len(self)
        while lo < hi:
            mid = (lo+hi)//2
            if seg[1] < self[mid][0]:
                hi = mid
            else:
                lo = mid+1
        # lo is just beyond last elt to be considered
        segs = []
        while first < lo:
            if Seg.overlap(self[first],seg):
                segs.append( Seg.intersection(self[first],seg) )
            first += 1
        return segs


    def overlap(self, seg):
        ovl = 0
        for s in self.intersection( seg ):
            ovl += Seg.len( Seg.intersection( s, seg ) )
        return ovl


    def getIndex(self,pt):
        # returns index to segment containing pt, or -1 if there's no such elt
        lo = 0
        hi = len(self)
        while lo < hi:
            mid = (lo+hi)//2
            if self[mid][1] <= pt:
                lo = mid+1
            else:
                hi = mid
        if lo==len(self):
            return -1
        if Seg.overlapelt(self[lo],pt):
            return lo
        return -1
        

    def loadView(self, paramDict, parentViews, viewDB, id):
        # Loads cache.  This base class implementation just copies the initialisation variables
        View.createView(self, paramDict, parentViews, viewDB, id)


    def init(self):
        # Initialisation.  Overridden init()'s are called by createView above
        pass


    def getFileName(self, name):
        # Helper function; returns unique filename to store cached data
        return self.viewDB.rootdict['cacheDir'] + self.__class__.__name__ + '-' + str(self.id) + '-' + name + '.cache'


    def cache(self,var,name):
        # Helper function; simple cache
        if True:
            filename = self.getFileName(name)
            handle = open(filename,'w+b')
            cPickle.dump(var, handle, self.viewDB.rootdict['binaryCache'])
            handle.close()
        else:
            self.cachedata.append( (var, name) )


    def _flushcache(self,write):
        if write:
            for var,name in self.cachedata:
                filename = self.getFileName(name)
                handle = open(filename,'w+b')
                cPickle.cump(var, handle, self.viewDB.rootdict['binaryCache'])
                handle.close()
        self.cache = []

            
    def uncache(self,name):
        # Helper function; simple cache
        filename = self.getFileName(name)
        handle = open(filename,'rb')
        obj = cPickle.load(handle)
        handle.close()
        return obj
        

newpath = os.path.dirname(os.path.abspath( __file__ )) + "/Views"
sys.path.append( newpath )
print newpath

import ChromosomeView
import AlignmentView
import AlignmentSegmentView
import AlignmentParView
import TabListView
import GeneListView
import MAFXParView
import RepeatView
import SegmentIndexView
import SegmentView
import SequenceSelectView
import TabListView
import EnsGeneView
import AnnotatorDataView


