import re, os, string




#
# Removes gaps from the gapped input sequence
# in SeqRecord format.  Returns a Seq.
#
def ungap(sequence):
    return re.sub('-','',sequence)





#
# Returns aligned sequences with double gaps removed
#
def removedoublegaps(seq1,seq2):
    idx = 0
    s1 = []
    s2 = []
    while idx < len(seq1):
        if seq1[idx]!='-' or seq2[idx]!='-':
            s1.append(seq1[idx])
            s2.append(seq2[idx])
        idx = idx + 1
    return ( ''.join(s1), ''.join(s2) )




#
# Returns subalignment, in coordinates of the first sequence
# Does not return leading unaligned flank, but does return trailing unaligned flank.
#
def subalignment(seq1,seq2,segm):
    cur = 0
    first = -1
    idx = 0
    while idx < len(seq1):
        if seq1[idx]!='-':
            if cur<=segm[0] and first == -1:
                first = idx
            if cur==segm[1]:
                return (seq1[first:idx],seq2[first:idx])
            cur += 1
        idx += 1
    if first != -1:
        return (seq1[first:],seq2[first:])
    return ("","")




#
# Returns list of ungapped alignments
# If  removerepeats == True, also cuts out lowercase sections
#
# When removing repeats, the number of 'gaps' may be larger than the number of gap sequences in glist
#
def ungappedalignments(seq1,seq2,returnidx = False):
    a1 = []
    a2 = []
    g = []
    alist = []
    glist = []
    align = False
    idx = 0
    plist1 = []
    plist2 = []
    p1 = 0
    p2 = 0
    while idx < len(seq1):
        if seq1[idx]!='-' or seq2[idx]!='-':
            if seq1[idx]!='-' and seq2[idx]!='-':
                if not align:
                    plist1.append(p1)    # record start position of this block
                    plist2.append(p2)
                    if g != []:
                        glist.append(''.join(g))
                        g = []
                a1.append(seq1[idx])     # add aligning column
                a2.append(seq2[idx])
                align = True
            else:
                if align:
                    alist.append((''.join(a1),''.join(a2)))    # record accumulated columns
                    a1 = []                                    # reset column lists
                    a2 = []
                align = False
                if seq1[idx]=='-':
                    g.append(seq2[idx])
                if seq2[idx]=='-':
                    g.append(seq1[idx])
        if seq1[idx] != '-':
            p1 += 1
        if seq2[idx] != '-':
            p2 += 1
        idx += 1
    if align:
        alist.append((''.join(a1),''.join(a2)))
    if g != []:
        glist.append(''.join(g))
    if returnidx:
        return (alist,plist1,plist2,glist)
    else:
        return alist




def ungappedalignments3(seq1,seq2,seq3,returnidx = False):
    a1 = []
    a2 = []
    a3 = []
    g = []
    alist = []
    glist = []
    align = False
    idx = 0
    plist1 = []
    plist2 = []
    plist3 = []
    p1 = 0
    p2 = 0
    p3 = 0
    while idx < len(seq1):
        if seq1[idx]!='-' or seq2[idx]!='-' or seq3[idx]!='-':
            if seq1[idx]!='-' and seq2[idx]!='-' and seq3[idx]!='-':
                if not align:
                    plist1.append(p1)
                    plist2.append(p2)
                    plist3.append(p3)
                    if g != []:
                        glist.append(''.join(g))
                        g = []
                    align = True
                a1.append(seq1[idx])
                a2.append(seq2[idx])
                a3.append(seq3[idx])
            else:
                if align:
                    alist.append((''.join(a1),''.join(a2),''.join(a3)))
                    a1 = []
                    a2 = []
                    a3 = []
                    align = False
                if seq1[idx]!='-':
                    g.append(seq1[idx])
                elif seq2[idx]!='-':
                    g.append(seq2[idx])
                else:
                    g.append(seq3[idx])
        p1 += (seq1[idx] != '-')
        p2 += (seq2[idx] != '-')
        p3 += (seq3[idx] != '-')
        idx += 1
    if align:
        alist.append((''.join(a1),''.join(a2),''.join(a3)))
    if g != []:
        glist.append(''.join(g))
    if returnidx:
        return (alist,plist1,plist2,plist3,glist)
    else:
        return alist





#
# Returns number of identical columns, and number of aligned columns
#
def alignidentity(seq1,seq2):
    idx = 0
    tot = 0
    id = 0
    while idx < len(seq1):
        if seq1[idx]!='-' and seq2[idx]!='-':
            tot += 1
            if seq1[idx].upper()==seq2[idx].upper():
                id += 1
        idx += 1
    return ( id, tot )




def alignidentityCG(seq1,seq2):
    idx = 0
    tot = 0
    id = 0
    totcg = 0
    idcg = 0
    tottv = 0      # total considered
    idtv = 0       # non-transversions
    ti = {'A':'G','G':'A','C':'T','T':'C','M':'x','R':'x','N':'x'}
    sq1 = seq1.upper()
    sq2 = seq2.upper()
    while idx < len(seq1):
        s1 = sq1[idx]
        s2 = sq2[idx]
        if s1!='-' and s2!='-':
            # See if this is a CpG site
            cg = 1
            if idx>0:
                dup1 = sq1[idx-1] + s1
                dup2 = sq2[idx-1] + s2
                if dup1 == "CA" or dup1 == "CG" or \
                   dup2 == "CA" or dup2 == "CG":
                    cg = 0
            if idx < len(seq1)-1:
                dup1 = s1 + sq1[idx+1]
                dup2 = s2 + sq2[idx+1]
                if dup1 == "TG" or dup1 == "CG" or \
                   dup2 == "TG" or dup2 == "CG":
                    cg = 0
            # Add counts
            totcg += cg
            tot += 1
            if s1 == s2:
                idcg += cg
                id += 1
            # now look at transversions
            tottv += 1
            if s1 != s2:
                if ti[s1] == s2:
                    # only a transition
                    idtv += 1
            else:
                # identical
                idtv += 1
        idx += 1
    return ( id, tot, idcg, totcg, idtv, tottv )



#
# Returns number of mutations on each branch; number of unidentifiable mutations;
# and total number of sites examined.  Returns same numbers but with possible
# CpG-related mutations removed.  All using parsimony.
#

def alignidentity3(seq0,seq1,seq2):
    sq0 = seq0.upper()
    sq1 = seq1.upper()
    sq2 = seq2.upper()
    mc0,mc1,mc2,mc012 = (0,0,0,0)
    mccg0,mccg1,mccg2,mccg012 = (0,0,0,0)
    idx = 0
    m0,m1,m2,m012 = 0,0,0,0
    totcg = tot = 0
    while idx < len(seq1):
        s0 = sq0[idx]
        s1 = sq1[idx]
        s2 = sq2[idx]
        if s1!='-' and s2!='-' and s0!='-':
            eq01 = s0 == s1
            eq12 = s1 == s2
            eq20 = s2 == s0
            cg = 1
            if idx>0 and ( (s0 in 'GA' and (sq0[idx-1] == 'C')) or \
                           (s1 in 'GA' and (sq1[idx-1] == 'C')) or \
                           (s2 in 'GA' and (sq2[idx-1] == 'C')) ):
                cg = 0
            if idx<len(seq1)-1 and ( (s0 in 'CT' and (sq0[idx+1] == 'G')) or \
                                     (s1 in 'CT' and (sq1[idx+1] == 'G')) or \
                                     (s2 in 'CT' and (sq2[idx+1] == 'G')) ):
                cg = 0
            if (not eq01) and (not eq12) and (not eq20):
                # "we're all different"
                mc012 += 1
                mccg012 += cg
            elif not (eq01 and eq12 and eq20):
                # "(I'm not)"
                if eq01:
                    mc2 += 1
                    mccg2 += cg
                elif eq12:
                    mc0 += 1
                    mccg0 += cg
                elif eq20:
                    mc1 += 1
                    mccg1 += cg
            else:
                # All the same
                pass
            tot += 1
            totcg += cg
        idx += 1
    return (mc0,mc1,mc2,mc012,tot,mccg0,mccg1,mccg2,mccg012,totcg)




#
# Returns number of mutations on each branch; number of unidentifiable mutations;
# and total number of sites examined.  Returns same numbers but with possible
# CpG-related mutations removed.  All using parsimony.
#

def alignidentity3tv(seq0,seq1,seq2):
    sq0 = seq0.upper()
    sq1 = seq1.upper()
    sq2 = seq2.upper()
    tot = 0
    tvmap = {'A':0,'G':0,'T':1,'C':1}
    tvcode = [3,2,1,0,0,1,2,3]   # maps binary code to parsimonious sequence (or identity)
    counts = [0,0,0,0]           # unique to seq0, seq1, seq2; identity

    idx = 0
    while idx < len(seq1):
        s0 = sq0[idx]
        s1 = sq1[idx]
        s2 = sq2[idx]
        # only look at full columns; and ignore N's, M'r and R's
        if s0 in "ACGT" and s1 in "ACGT" and s2 in "ACGT":
            counts[ tvcode[ tvmap[s2] + 2*tvmap[s1] + 4*tvmap[s0] ] ] += 1
            tot += 1
        idx += 1
    return (counts[0],counts[1],counts[2],tot)





#
# Returns all aligned columns in Seq1 and Seq2, and percentage identity
#
def alignonly(seq1,seq2):
    idx = 0
    s1 = []
    s2 = []
    pic = 0
    while idx < len(seq1):
        if seq1[idx]!='-' and seq2[idx]!='-':
            s1.append(seq1[idx])
            s2.append(seq2[idx])
        idx = idx + 1
    return ( ''.join(s1), ''.join(s2) )




def reverse(seq):
    map = {'a':'t','t':'a','c':'g','g':'c','A':'T','T':'A','C':'G','G':'C','N':'N','n':'n','-':'-','u':'u'}
    try:
        rseq = [ map[c] for c in seq ]
        rseq.reverse()
        return ''.join(rseq)
    except:
        return None


    



#
# Run dust on a sequence
#
def dust(seq,cutoff = 28):
    dustfilename = ".dustfile"
    file = open(dustfilename,'w')
    file.write(">1\n" + seq + "\n")
    file.close()
    output = os.popen("dust " + dustfilename + " " + str(cutoff))
    lines = output.readlines()
    output.close()
    outseq = ""
    for i in range(1,len(lines)):
        outseq += lines[i][:-1]
    return outseq
        



#
# Checks for low complexity (true if single N in dust output)
#
def lowcomplexity(seq):
    return (re.search('N',dust(seq)) != None)





#
# Returns list of between-indel (ungapped) lengths
#
def ungappedlengths(seq1,seq2):
    uglist = []
    gap = 1
    gaplen = 0
    for idx in range(len(seq1)):
        ch1 = seq1[idx]
        ch2 = seq2[idx]
        if ch1 == '-' or ch2 == '-':
            if not gap:
                gap = 1
                gaplen = 1
            else:
                gaplen = gaplen+1
        else:
            if gap:
                gap = 0
                if gaplen>0:
                    uglist.append( gaplen )
    if gap:
        uglist.append( gaplen )
    return uglist




#
# Returns c+g and nucleotide count
#
def cplusg(seq):
    cg = 0
    nuc = 0
    seq = seq.upper()
    for idx in range(len(seq)):
        if seq[idx]=='C' or seq[idx]=='G':
            cg += 1
            nuc += 1
        elif seq[idx]=='A' or seq[idx]=='T':
            nuc += 1
    return cg, nuc




def codon2aa(codon):
    return {
        'GCA':'A',
        'GCC':'A',
        'GCG':'A',
        'GCT':'A',
        'TGC':'C',
        'TGT':'C',
        'GAC':'D',
        'GAT':'D',
        'GAA':'E',
        'GAG':'E',
        'TTC':'F',
        'TTT':'F',
        'GGA':'G',
        'GGC':'G',
        'GGG':'G',
        'GGT':'G',
        'CAC':'H',
        'CAT':'H',
        'ATA':'I',
        'ATC':'I',
        'ATT':'I',
        'AAA':'K',
        'AAG':'K',
        'TTA':'L',
        'TTG':'L',
        'CTA':'L',
        'CTC':'L',
        'CTG':'L',
        'CTT':'L',
        'ATG':'M',
        'AAC':'N',
        'AAT':'N',
        'CCA':'P',
        'CCC':'P',
        'CCG':'P',
        'CCT':'P',
        'CAA':'Q',
        'CAG':'Q',
        'AGA':'R',
        'AGG':'R',
        'CGA':'R',
        'CGC':'R',
        'CGG':'R',
        'CGT':'R',
        'AGC':'S',
        'AGT':'S',
        'TCA':'S',
        'TCC':'S',
        'TCG':'S',
        'TCT':'S',
        'ACA':'T',
        'ACC':'T',
        'ACG':'T',
        'ACT':'T',
        'GTA':'V',
        'GTC':'V',
        'GTG':'V',
        'GTT':'V',
        'TGG':'W',
        'TAC':'Y',
        'TAT':'Y',
        'TAA':'X',
        'TGA':'X',
        'TAG':'X'}[codon];


    
