#
# Fit neutral indel model to histogram[rows,column]
# rows = range, column = number
#
fitnim <- function( histograms, rows, column )
{
  data = cbind( histograms[ rows, 1], log(histograms[ rows, column ] + 0.01) )   # this may produce negative numbers
  colnames(data) = c("igslen", "logcount")
  w = sqrt( data[,2] )
  w[ is.na(w) ] = 0.00000001
  fit = lm( logcount ~ igslen, as.data.frame(data), weights = w ) 
  return ( fit )
}

#
# Plot histogram (in log coords) and the NIM fit
# Row ranges need not coincide with that for linear model
#
plotnim <- function( histograms, rows, column, fit )
{
  data = cbind( histograms[ rows, 1], log(histograms[ rows, column ]) )
  xmax = data[ dim(data)[1] ]
  intercept = coef(fit)[1]
  ymax = intercept + xmax * coef(fit)[2]
  if (par("yaxs")=="d") {                                 # simulate "d" style - keep old limits
    margin = 0.04*(par("usr")[4]-par("usr")[3])/1.08      # compute 4% margin
    ylim = c(par("usr")[3]+margin,par("usr")[4]-margin)   # recover old ylim
    par(yaxs="r")
    plot(data,ylim=ylim)
    par(yaxs="d")
  } else {
    plot(data)
  }
  lines( c(0,xmax), c(intercept, ymax) )
}

#
# Compute relative weighted deviance, measure of unexplained signal
#
dev <- function( histograms, rows, column, fit )
{
  data = cbind( histograms[ rows, 1], log(histograms[ rows, column ] + 0.01) )
  w = sqrt( data[,2] )
  w[ is.na(w) ] = 0.00000001
  d = (w * residuals(fit)) %*% residuals(fit)
  intercept = coef(fit)[1]
  rc = coef(fit)[2]
  avg = intercept + (data[1,1] + data[length(data[,1]),1])*rc/2
  v = (w * (data[,2]-avg)) %*% (data[,2]-avg) + 1e-12
  return ( d/v )
}

#
# Compute rows that minimize (a weighted version of) the deviance per independent observable
#
bestrows <- function( histograms, starts, ends, column ) 
{
  interv = c(0,0)
  mindev = 99999999
  for (x in starts) {
    for (y in ends) {
      rows = c(x:y)
      fit = fitnim( histograms, rows, column )
      #
      # compute the deviance (total variance) per independent column
      #
      d = dev( histograms, rows, column, fit ) #/ ((y - x - 1))
      if (d < mindev) {
	 mindev = d
	 interv = c(x:y)
      }
    }
  }
  return (interv)
}

#
# Compute upper and lower bounds on the number of indel-purified nucleotides
# Also computes the total number of nucs in indel-purified segments (including neutral overhang),
# and total number of aligning nucleotides.
#
# There's code-repetition here; this function operators on 'fit', whereas predneutral, predips and ipssummary
# compute much the same things, but operate on 'analysis'.  Sorry.
#
conserved <- function( histograms, firstrow, column, fit )
{
  intercept = coef(fit)[1]
  rc = coef(fit)[2]
  response = exp(histograms[,1] * rc + intercept)
  avgindellen = 1 / (1-exp(rc))

  purified_low = histograms[,1] - 2*avgindellen
  purified_high = histograms[,1] - avgindellen
  purified_toohigh = histograms[,1]

  histlen = dim(histograms)[1]
  subset = 1:histlen
  subset[subset < firstrow] = 0
  subset[subset >= firstrow] = 1

  lowerbound = ((histograms[ ,column ] - response) * purified_low) %*%  subset
  upperbound = ((histograms[ ,column ] - response) * purified_high) %*% subset
  nobound = ((histograms[ ,column ] - response) * purified_toohigh) %*% subset
  total = histograms[ ,column ] %*% histograms[ ,1 ]
  neutral = (response * histograms[,1]) %*% subset

  #print(paste("row=",firstrow," neutral=",response[firstrow]," hist=",histograms[firstrow,column]))
  return ( c(lowerbound,upperbound,nobound,total,neutral) )
}



#
# predicted amount of neutral entries (igs) at every row
#
predneutral_col <- function( histograms, analysis, column )
{
  a = analysis$a[[ column - analysis$coloffset ]]       
  b = analysis$b[[ column - analysis$coloffset ]]
  return( exp(a + histograms[,1]*b) )
}

#
# predicted neutral version of histogram
#
predneutral_hist <- function( histograms, analysis )
{
  h = histograms[,1]
  for (colno in analysis$columns) {
    h = cbind(h, predneutral_col( histograms, analysis, colno ) )
  }
  return(h)
}


#
# predicted amount of neutral nucleotides at and above firstrow
#
predneutral <- function( histograms, analysis, firstrow, column )
{
  n0 = histograms[ firstrow, 1 ]
  a = analysis$a[[ column - analysis$coloffset ]]       
  b = analysis$b[[ column - analysis$coloffset ]]
  # sum of n exp(a + b n), for n from n0 to infinity
  return( exp(a + n0*b) * (n0 - exp(b)*(n0-1) ) / (1-exp(b))^2 )
}


#
# predicted amount of ungapped segments of nucleotides at and above firstrow
# This includes IPS and neutral segments
# can also calculate how many of those are exons, using other columns in the histogram table
#
predcons <- function( histograms, analysis, firstrow, column, exoncol = 0)
{
  histlen = dim(histograms)[1]
  subset = 1:histlen
  subset[subset < firstrow] = 0
  subset[subset >= firstrow] = 1
  return( as.numeric((histograms[,1] * subset) %*% histograms[,column + exoncol]) )
}


#
# predicted amount of ips nucleotides at and above firstrow
# Bound = 0: all IPS above thresholds
# Bound = 1: upper bound of nucleotides under selection
# Bound = 2: lower bound of nucleotides under selection
#
predips <- function( histograms, analysis, firstrow, column, bound = 0 )
{
  histlen = dim(histograms)[1]
  subset = 1:histlen
  subset[subset < firstrow] = 0
  subset[subset >= firstrow] = 1
  avgindellen = 1 / analysis$gapprobability[[ column - analysis$coloffset ]]
  response = predneutral_col( histograms, analysis, column )

  return( as.numeric(((histograms[,1] - bound*avgindellen) * subset) %*% (histograms[,column] - response)) )
}



#
# compute: lower bound conserved, upper bound conserved, total aligned, 
# and: indel rate, upper and lower proportions, and regression segment, per GC
#
analyze <- function( histograms, columns, startrows=10:30, endrows=40:110, exoncol = 0, plotfile = FALSE, fixedinterval = FALSE )
{
  totlower = 0
  totupper = 0
  totall = 0
  totalign = 0
  totexon = 0
  propupper_gc = NULL
  proplower_gc = NULL
  propall_gc = NULL
  gapprobability_gc = NULL
  a_gc = NULL
  b_gc = NULL
  uplimits_gc = NULL
  lowlimits_gc = NULL
  if (plotfile != FALSE) {
    postscript( file = plotfile )
    if (length(columns) < 4) {
      par(mfrow = c(1,length(columns)))
    } else if (length(columns) == 4) {
      par(mfrow = c(2,2))
    } else {
      par(mfrow = c(4,1+floor((length(columns)-1)/4)))
    }
    par(ann=FALSE)                   # do not print titles etc.
    par(cex=0.2)                     # define symbol/text size
    par(cex.axis=2)                  # relative axis annotation size
    par(cex.lab=2)                   # relative label size
    par(cex.main=3)                  # relative main title size
    par(mar=c(2.5,2.5,2.0,2.5))      # margins in cex units
    par(mgp=c(-6,-2,0))              # axis title, label, line (location)
    par(omd=c(0.03,0.90,0.03,0.97))  # outer margins, relative
  }
  for (col in columns) {
    if (fixedinterval) {
      sr = startrows[ col-1 ]
      er = endrows[ col-1 ]
    } else {
      sr = startrows
      er = endrows
    }
    rows = bestrows( histograms, sr, er, col )
    print( c(col, rows[[1]], rows[[ length(rows) ]]) )
    fit = fitnim( histograms, rows, col )
    if (plotfile != FALSE) { 
      # exclude intergap-length=1, because counts are low and distort the plot
      plotnim(histograms, c(3:endrows[ length(endrows) ]), col, fit)
      intercept = coef(fit)[1]
      par(lty=3)                     # dotted line
      lines( c( rows[[1]], rows[[1]] ), c( 0, intercept ) )
      lines( c( rows[[length(rows)]], rows[[length(rows)]] ), c( 0, intercept ) )
      par(lty=1)
      title(paste("GC bin",col-1))
      par(yaxs="d")                 # do not change y axes
    }
    gapprobability = 1 - exp( coef(fit)[2] )
    names(gapprobability) = NULL
    limits = c(rows[1], rows[ length(rows) ])
    midrow = (limits[1]+limits[2])/2
    cons = conserved( histograms, midrow, col, fit )
    totlower = totlower + cons[1]
    totupper = totupper + cons[2]
    totall = totall + cons[3]
    totalign = totalign + cons[4]
    proplower_gc = c(proplower_gc, cons[1]/cons[4] )
    propupper_gc = c(propupper_gc, cons[2]/cons[4] )
    propall_gc = c(propall_gc, cons[3]/cons[4] )
    gapprobability_gc = c(gapprobability_gc, gapprobability)
    lowlimits_gc = c(lowlimits_gc, limits[[1]])
    uplimits_gc = c(uplimits_gc, limits[[2]])
    a_gc = c(a_gc, coef(fit)[1] )
    b_gc = c(b_gc, coef(fit)[2] )
    totexon = totexon + predcons( histograms, analysis, 1, col, exoncol )
  }
  if (plotfile != FALSE) { graphics.off() }
  return (list( totlower = totlower, 
		totupper = totupper, 
		totips = totall, 
		totalign = totalign, 
		totexon = totexon,
                gapprobability = gapprobability_gc, 
		proplower = proplower_gc, 
		propupper = propupper_gc, 
		propips = propall_gc,
                lowlimits = lowlimits_gc, 
		uplimits = uplimits_gc, 
		a = as.numeric(a_gc), 
		b = as.numeric(b_gc), 
		columns = columns, 
		coloffset = columns[[1]]-1 ) )
}



##############################################################################################
#
# optimal cutoffs for a fixed FDR and maximum predicted amount of ips nucleotides
#
#
#

# Helper functions
#smooth_histogram <- function( histogram, idx, b, width )
#{
#    sm = 0
#    width = 0
#    for (i in -width:width) {
#      sm = sm + histogram[idx+i]*exp(-b*i)
#    }
#    return (sm/(2*width+1))
#}
#
#interpolate_histogram <- function( histograms, idx, b, width=1 )
#{
#  flr = floor(idx)
#  cei = array( c( rep(1, times=length(idx[,1])), rep(0, times=length(idx[,1]) ) ), dim=dim(idx) ) + flr
#  coeff = (cei - idx)[,1]
#  return( smooth_histogram(histograms,flr,b,width)*coeff + smooth_histogram(histograms,cei,b,width)*(1-coeff) )
#}

smooth_row <- function( histogram, b )
{
  h1 = histogram[c(1,1:length(histogram)-1)]
  h2 = histogram[c(2:length(histogram),length(histogram))]
  return( (histogram + 0.5*exp(b)*h1 + 0.5*exp(-b)*h2 )/2 )
}

# Calculate Lagrange cutoffs (row numbers) for a given 'multiplier' alpha
cutoffs <- function( histograms, analysis, alpha, width=1 )
{
  a = analysis$a
  b = analysis$b
  row1 = histograms[[1,1]]
  cols = analysis$columns
  coloffset = analysis$coloffset
  n = NULL

  for (col in cols) {
    ratios = smooth_row(histograms[,col], b[col - coloffset ]) / predneutral_col(histograms,analysis,col)
    i = 20
    while (i<=length(ratios) && ratios[[i]]<alpha) {
      i = i + 1
    }
    if (i<=length(ratios)) {
      dratio = ratios[[i]] - ratios[[i-1]]
    } else {
      dratio = 10000
    }
    n = append(n, i-1+(alpha-ratios[[i-1]])/dratio )
  }
    
  return (as.numeric(n))
}

#
# total predicted amounts of neutral and ips sequence, and FDR, for a vector of cutoffs
#
ipssummary <- function( histograms, analysis, cutoffs, exoncol = 0 )
{
  totips = 0
  totsel_up = 0
  totsel_low = 0
  totneut = 0
  totexon = 0
  for (gc in 1:length(analysis$columns)) {

    totneut = totneut + predneutral( histograms, analysis, cutoffs[[gc]], analysis$columns[[gc]] )
    totips = totips + predips( histograms, analysis, cutoffs[[gc]], analysis$columns[[gc]], 0 )
    totsel_up = totsel_up + predips( histograms, analysis, cutoffs[[gc]], analysis$columns[[gc]], 1 )
    totsel_low = totsel_low + predips( histograms, analysis, cutoffs[[gc]], analysis$columns[[gc]], 2 )
    totexon = totexon + predcons( histograms, analysis, cutoffs[[gc]], analysis$columns[[gc]], exoncol )

    if (length(totneut) == 0) { totneut = 0 }
    if (length(totexon) == 0) { totexon = 0 }
    if (length(totips) == 0) { totips = 0 }
    if (length(totsel_up) == 0) { totsel_up = 0 }
    if (length(totsel_low) == 0) { totsel_low = 0 }

  }
  return( c(totips, totneut, totneut/(totips+totneut), totexon, totsel_up, totsel_low ) )
}

#
# find Lagrange multiplier alpha, and cutoffs (as lengths, not columns), for a given FDR
# exoncol = offset of columns with exon density w.r.t. IPS histogram counts (3 * 20)
#
fdrcutoffs <- function( histograms, analysis, fdr, exoncol = 0, width=1)
{
  minalpha = 1
  maxalpha = 100

  minn = cutoffs( histograms, analysis, minalpha, width )
  maxn = cutoffs( histograms, analysis, maxalpha, width )

  minfdr = ipssummary( histograms, analysis, minn )[[3]]
  maxfdr = ipssummary( histograms, analysis, maxn )[[3]]

  while (maxalpha - minalpha > 0.01) {

    alpha = sqrt( minalpha * maxalpha )
    n = cutoffs( histograms, analysis, alpha, width )
    fdr0 = ipssummary( histograms, analysis, n )[[3]]

    if (is.na(fdr0)) {
      fdr0 = fdr;
    }

    if( fdr0 > fdr ) {
      minalpha = alpha
      minn = n
      minfdr = fdr0
    } else {
      maxalpha = alpha
      maxn = n
      maxfdr = fdr0
    }
  }

  cutoffs = floor( (1 + maxn + minn)/2 )
  row0 = histograms[[1,1]]
  cutoffs = cutoffs - 1 + row0

  summary = ipssummary( histograms, analysis, cutoffs, exoncol )

  return ( list( cutoffs = cutoffs, 
		 totcons = summary[[1]]+summary[[2]],
                 totips = summary[[1]], 
		 totsel_upper = summary[[5]],
		 totsel_lower = summary[[6]],
                 totneutral = summary[[2]],
		 fdr = summary[[3]],
		 totexon = summary[[4]] ) )

}

