#
# A set of scripts to analyze the gap distribution in genomic alignments, and
# in particular, for identifying Indel-Purified Segments
#

from __future__ import generators
import sys, random, os, re, math, copy
import rpy

newpath = os.path.dirname(os.path.abspath( __file__ )) + "/DBView"
sys.path.append(newpath)

import dbview
import aligntools
from mygetopt import *
from segviewtools import *

try:
    import psyco
except:
    print "(Psyco not available - will work, but slower)"

exoncol = 3

debug = False


#
# subroutines
#



def dopsyco():
    try:
        psyco.profile(0.09,memory=100)
        psyco.profile(0.2,memory=250)
    except:
        pass
    


# dump all settings
def dumpsettings():
    if settings['verbose'] > 0:
        print "Parameter settings:-"
        print
        for id in settings:
            print "%25s = %s" % (id,settings[id])
        print


# Returns list of segments representing the known (non-N) nucleotides
def filterN(seq):
    cnt = seq.count('N')
    if cnt == 0:
        return [(0,len(seq))]
    seqs = []
    if cnt == len(seq):
        return seqs
    l = -1
    for i in range(0,len(seq)):
        if seq[i]!='N':
            if l == -1:
                l = i
        else:
            if l != -1:
                seqs.append((l,i))
                l = -1
    if l != -1:
        seqs.append((l,len(seq)))
    return seqs



#
# This writes out a histogram, with implicit labels determined by binsize
# It can also accumulate histograms in histdata, to be written to a file later
#
def writehistogram( hist, binsize, label = "bin", countlabel = "count", histdata = False, name='outputfile' ):
    if type(histdata) == type([]):
        # multiple column
        if binsize == 0:
            # write the list
            file = open(settings[name],'a')
            formatstring = ("%s\t" * len(histdata[0]))[:-1] + "\n"
            for line in histdata:
                file.write( formatstring % tuple(line) )
            file.close()
            return
        # add to the list
        if histdata == []:
            # first column
            histdata = [ [label] ]
            for line in range(len(hist)):
                histdata.append( [line * binsize] )
        histdata[0].append( countlabel )
        for line in range(len(hist)):
            histdata[line+1].append( hist[line] )
        return histdata
    else:
        # single column
        file = open(settings[name],'a')
        file.write("%s\t%s\n" % (label, countlabel) )
        for line in range(len(hist)):
            file.write("%s\t%s\n" % (line * binsize, hist[line]))
        file.close()



def writedicthistogram( hist, name='' ):
    file = open(settings[name],'w')
    keys = hist.keys()
    keys.sort()
    for k in keys:
        file.write(k)
        for e in hist[k]:
            file.write('\t%s' % e)
        file.write('\n')
    file.close()
    
    

# build dictionary mapping loci (in kilobases) on species 1, onto trusted homologous (chrom, locus)
# pairs within species 2, to filter out suspicious alignments that are not in conserved synteny with
# neighbouring sequence
#
# 1: include srep into segtrepv
def syntenyanchors( alignmentparv, segtrepv, segrepeatv ):

    syndict = {}
    segtrepinvidxv = inverseidxview( db, segtrepv )
    segrepeatinvidxv = inverseidxview( db, segrepeatv )
    badchroms = {}

    curchromstrand = ""
    curtotlen = 0
    curpos = -1

    for ap in alignmentparv:

        # get segment
        start = ap[ alignmentparv.fields['start1'] ]
        end = ap[ alignmentparv.fields['end1'] ]
        if end<start:
            print "Error!"
            sys.exit(1)
        if ap[alignmentparv.fields['strand1']] == '+':
            seg = (start,end)
        else:
            seg = ( len(chromv)-end, len(chromv)-start )
            
        # remove repetitive elements
        seglist = segtrepinvidxv.intersection( seg )
        seglist = segrepeatinvidxv.intersection( seglist )
        
        # count the bits that contribute to the anchor
        totlen = 0
        for s in seglist:
            totlen += Seg.len(s)

        # add to whatever we already found
        start2 = ap[ alignmentparv.fields['start2'] ]
        end2 = ap[ alignmentparv.fields['end2'] ]  
        chr2 = ap[ alignmentparv.fields['chr2'] ]
        chrstrand2 = chr2 + ap[ alignmentparv.fields['strand2'] ]

        #if curchromstrand == chrstrand2 and abs( start2 - curpos ) < settings['maxsyntenygap']:
        #    curtotlen += totlen
        #else:
        #    curtotlen = totlen
        #curpos = end2
        curtotlen = totlen
        curchromstrand = chrstrand2

        # do we have a trustable anchor?
        if curtotlen > settings['minsynteny']:

            seg2 = (start2,end2)         # this is relative to the strand
                
            for i in range(int((seg[0]-500)/1000),int((seg[1]+500)/1000)):
                loc2 = seg2[0] + (seg2[1]-seg2[0]) * (i * 1000 - seg[0]) / (seg[1]-seg[0])
                loc2 = int( (loc2 + 500)/1000 )
                if syndict.has_key(i):
                    if (chrstrand2,loc2) not in syndict[i]:
                        syndict[i].append((chrstrand2,loc2))
                else:
                    syndict[i] = [(chrstrand2,loc2)]
        else:
            pass

    return syndict


    


# build dictionary mapping loci (in kilobases) on species 1, onto trusted homologous (chrom, locus)
# pairs within species 2, to filter out suspicious alignments that are not in conserved synteny with
# neighbouring sequence
def genesyntenyanchors( alignmentparv, segexonv, genev, segtranscriptv ):

    syndict = {}

    covered = {}

    for ap in alignmentparv:

        # get segment
        start = ap[ alignmentparv.fields['start1'] ]
        end = ap[ alignmentparv.fields['end1'] ]
        if end<start:
            print "Error!"
            sys.exit(1)
        if ap[alignmentparv.fields['strand1']] == '+':
            seg = (start,end)
        else:
            seg = ( len(chromv)-end, len(chromv)-start )
        covered[ seg[0] // 1000 ] = 1

        if segexonv.overlap( seg ) > 25:
            # find corresponding gene to generate anchors

            geneidx = segtranscriptv.getNearestIndex( (seg[0]+seg[1])/2 )
            if geneidx == -1: continue

            gene = genev[geneidx]
            codingstart = gene[ genev.fields['codingstart'] ]
            codingend = gene[ genev.fields['codingend'] ]

            if Seg.overlap( seg, (codingstart,codingend) ) > 25:
                # segment overlaps coding exon - so trust it

                start2 = ap[ alignmentparv.fields['start2'] ]
                end2 = ap[ alignmentparv.fields['end2'] ]  
                chr2 = ap[ alignmentparv.fields['chr2'] ]
                chrstrand2 = chr2 + ap[ alignmentparv.fields['strand2'] ]
                seg2 = (start2,end2)         # this is relative to the strand

                # drop anchors around this segment
                for i in range(int((seg[0]-500)/1000),int((seg[1]+500)/1000)):
                    loc2 = int( seg2[0] + (seg2[1]-seg2[0]) * (i * 1000 - seg[0]) / (seg[1]-seg[0]) )
                    loc2 = int( (loc2 + 500)/1000 )
                    if syndict.has_key(i):
                        if (chrstrand2,loc2) not in syndict[i]:
                            syndict[i].append((chrstrand2,loc2))
                    else:
                        syndict[i] = [(chrstrand2,loc2)]

    # now see how much of aligned material is in conserved synteny with something
    covlist = covered.keys()
    covlist.sort()
    covidx = 0

    synlist = syndict.keys()
    synlist.append(3000000)
    synlist.sort()
    synidx = 0

    syntenous = 0
    nonsyntenousintrons = 0
    
    while covidx < len(covlist):
        cov = covlist[covidx]
        while abs( synlist[synidx] - cov ) > abs( synlist[synidx+1] - cov ):
            synidx += 1
        dist = abs(synlist[synidx]-cov)
        if dist < settings['synwindow']//4000:
            syntenous += 1
        else:
            nonsyntenousintrons += segtranscriptv.overlap( (cov*1000,(cov+1)*1000) )
        covidx += 1
        
    print "Syntenous: ",syntenous," of ",len(covlist), "(",(syntenous*100.0)/len(covlist),"%);",\
          "non-synt introns=",nonsyntenousintrons,"(",(nonsyntenousintrons*100.0)/( segtranscriptv.overlap( Seg.allSeg ) )," %)"

    return syndict






#
# See if an alignment is in conserved synteny with anchors
#
def issyntenous( pos1, pos2, chr2, strand2, syndict ):
    quartrange = settings['synwindow'] / 4000
    h = int((pos1+500)/1000)       # position onto base chromosome
    m = int((pos2+500)/1000)
    for i in range(h,h+quartrange):     # properly aligning within 100 kb window?
        if syndict.has_key(i+1):
            jlist = syndict[i+1]
            for (chrstrand,j) in jlist:
                if chrstrand==chr2+strand2 and (m-quartrange < j < m+quartrange):
                    return True
        if syndict.has_key(2*h-i):
            jlist = syndict[2*h-i]
            for (chrstrand,j) in jlist:
                if chrstrand==chr2+strand2 and (m-quartrange < j < m+quartrange):
                    return True

    return False



                      

#
# Open all chromosomes of a species
#
def getchromosomes(directory):
    f = os.popen("ls -1 --color=never "+directory+"*")                                                 # get listing of all files
    extension = ".fa"
    chroms = [ c[:-len(extension)-1] for c in f.readlines() if c[-len(extension)-1:-1] == extension]   # ignore '\n' at end
    if len(chroms)<21:
        print "Found only these chromosomes:",chroms
    chromdict = {}
    for chr in chroms:
        chromv = db.open("ChromosomeView",{'filename':directory + '/' + chr + extension})
        chromdict[chr] = chromv
    return chromdict



#
# Progress indicator
#
class Progress(object):

    def progress( self, current, end ):

        currenttime = os.times()[-1]
        if current == 0:
            self.start = currenttime
            self.lasttime = currenttime

        if self.lasttime > currenttime - 3.0:
            return
        self.lasttime = currenttime
            
        elapsed = currenttime - self.start
        percent = current/(end + 0.0)
        projected = elapsed / (percent+0.0001)

        if settings['verbose'] >= 2:
            print "Processing: %s of %s (%s %%; elapsed=%d s; estimated: %d s)\r" % \
                  (current,end,int(percent*100.0), int(elapsed), int(projected) ),
            sys.stdout.flush()



#
# Returns total number of aligning nucleotides (including small gaps)
#
def aligning( alignmentparv ):

    alignable = 0
    for idx in range( len(alignmentparv) ):

        # get coordinates of alignment
        alignpar = alignmentparv[idx]
        start1 = alignmentparv[idx][ alignmentparv.fields['start1'] ]
        end1 = alignmentparv[idx][ alignmentparv.fields['end1'] ]

        alignable += end1 - start1

    return alignable



#
# Returns a list of (overlap, class, name) triplets
#
def repeatoverlap( segment, segtev, tev ):

    rolist = []
    overlapidx = segtev.indexIntersection( segment )
    for idx in overlapidx:
        repeat = tev[idx]
        repseg = (repeat[ tev.fields['start'] ], repeat[ tev.fields['end'] ])
        repcls = repeat[ tev.fields['class'] ]
        repnam = repeat[ tev.fields['repeat'] ]
        rolist.append( (Seg.len( Seg.intersection( segment, repseg ) ), repcls, repnam ) )
    return rolist




#
# Compute local neutral rates, by adding local counts until some threshold is reached
#
def neutralid( ratev, rateidxv, ratetots, position, idlabel = 'tvid', totlabel = 'tvtot' ):

    global _nid_lastseg, _nid_lastrate

    try:
        if Seg.overlapelt( _nid_lastseg, position ):
            return _nid_lastrate
    except NameError:
        pass

    maxdist = 2       # number of windows around current site that we consider
    mincount = 1500   # minimum number of contributing sites before estimate is trusted

    totacc = 0
    idacc = 0

    idx = rateidxv.getNearestIndex( position )
    _nid_lastseg = (ratev[idx][ ratev.fields['start'] ], ratev[idx][ ratev.fields['end'] ])

    for d in range(maxdist):

        if idx-d >= 0:

            id = ratev[idx-d][ ratev.fields[idlabel] ]
            tot = ratev[idx-d][ ratev.fields[totlabel] ]

            idacc += id
            totacc += tot

            if totacc > mincount:
                break

        if idx+d < len(ratev) and d>0:

            id = ratev[idx+d][ ratev.fields[idlabel] ]
            tot = ratev[idx+d][ ratev.fields[totlabel] ]

            idacc += id
            totacc += tot

            if totacc > mincount:
                break

    if totacc <= mincount:

        # use averages
        totacc = ratetots[ ratev.fields[totlabel] ]
        idacc = ratetots[ ratev.fields[idlabel] ]

    rate = idacc / (0.0+totacc)
    _nid_lastrate = rate

    return rate




def calcbestFDR( length, gc ):
    global _thresholds
    try:
        thresholds = _thresholds
    except:
        thresholds = eval( open(settings['thresholdfile'],'r').readline() )
        _thresholds = thresholds

    bestfdr = 1.0
    for fdr in thresholds:
        if length >= thresholds[fdr][ gc ]:
            bestfdr = min(bestfdr, fdr)
    return bestfdr



#
# Converts a sequence into a GC content class (usually from 0 to 20)
#
def getgcclass( gcseq ):
    gcseq = gcseq.upper()
    gcf = (0.05 + gcseq.count('C') + gcseq.count('G')) / (0.1 + len(gcseq) - gcseq.count('N'))
    for i in range( len( gcbins ) ):
        if gcf< gcbins[i]:
            return i,gcf
    return len(gcbins),gcf




def getgcclassfrompos( start, chromv ):
    gcflank = settings['gcwindowsize']/2
    posl = max( start - gcflank, 0)
    posr = min( start + gcflank, len(chromv))
    igswindow = chromv[ posl:posr ]
    gcclass, gcfraction = getgcclass( igswindow )
    return gcclass, gcfraction

            




##############################################################################################################################
#
#
# The main subroutines that do all the work
#
#
##############################################################################################################################



#
# Load data into global variables
#
def loaddata():
    global chromv, segknownv, segknownidxv, trfv, segtrepv, tev, segtev, genev, segexonv
    global apv, alignmentv, anchors, ratev, rateidxv, ratetots, segtranscriptv
    #global secondarychroms

    # open chromosome view
    chromv = db.open("ChromosomeView",
                     {'filename': settings['basechromdir'] + settings['chromosome'] + '.fa'},
                     [])
    
    # find the segments representing known nucleotides
    segknownv = db.open("SequenceSelectView",
                        {'criterion':'filterN',
                         'window':10000},
                        [chromv],
                        globals = globals() )
    segknownidxv = db.open("SegmentIndexView",{},[segknownv])

    # load tandem repeats (if available)
    if settings['tandemrepeatfile'] != "":
        trfv = db.open("TabListView",
                       {'filename':settings['tandemrepeatfile'],
                        'keys':['chr',1,settings['chromlabelprefix']+settings['chromosome'],
                                'start',2,0,
                                'end',3,0],
                        'cache':False})
        segtrepv = db.open("SegmentView",
                           {'singleSegment':(trfv.fields['start'],trfv.fields['end']),
                            'shoulder':0},
                           [trfv])
    else:
        trfv = None
        segtrepv = db.open("SegmentView",{},[])

    # load transposable elements (if available)
    if settings['repeatmaskfile'] != "":
        filename = settings['repeatmaskfile']
        if re.search("\%s",filename) != None:
            filename = filename % settings['chromosome']
        tev = db.open("RepeatView",
                      {'filename': filename,
                       'teonly':1})
        segtev = db.open("SegmentIndexView",{'singleSegment':(tev.fields['start'],tev.fields['end'])},[tev])
    else:
        tev = None
        segtev = db.open("SegmentIndexView",{},[])

    # load neutral rates (if available)
    if settings['localratefile'] != "":
        filename = settings['localratefile']
        print "Loading file ",filename," for chr=",settings['chromlabelprefix']+settings['chromosome']
        ratev = db.open("TabListView",
                        {'filename':filename,
                         'keys':['chr',0,settings['chromlabelprefix']+settings['chromosome'],
                                 'start',1,0,
                                 'end',2,0,
                                 'id',3,0,
                                 'tot',4,0,
                                 'cpgid',5,0,
                                 'cpgtot',6,0,
                                 'tvid',7,0,
                                 'tvtot',8,0],
                         'skiplines':1,
                         'cache':False})
        print "Length: ",len(ratev)
        if len(ratev) == 0:
            ratev = None
        else:
            rateidxv = db.open("SegmentIndexView",
                               {'singleSegment':(ratev.fields['start'],ratev.fields['end']),
                                'shoulder':0},
                               [ratev])
            ratetots = ratev[0][:]            # chrom, start
            for idx in range(1,len(ratev)):
                ratetots[2] = ratev[idx][2]   # end
                ratetots[3] += ratev[idx][3]
                ratetots[4] += ratev[idx][4]
                ratetots[5] += ratev[idx][5]
                ratetots[6] += ratev[idx][6]
                ratetots[7] += ratev[idx][7]
                ratetots[8] += ratev[idx][8]
    else:
        ratev = None

    # load exons (if available)
    if settings['genefile'] != "":
        filename = settings['genefile']
        if re.search("\%s", filename) != None:
            filename = filename % settings['chromosome']
        genev = db.open("EnsGeneView",
                        {'filename':filename,
                         'chromosome':settings['chromlabelprefix']+settings['chromosome']})
        segexonv = db.open("SegmentIndexView",
                           {'segmentList':genev.fields['exons'],
                            'shoulder':0},[genev])
        segtranscriptv = db.open("SegmentIndexView",
                                 {'singleSegment':(genev.fields['start'],genev.fields['end']),
                                  'shoulder':0},[genev])
    else:
        segexonv = db.open("SegmentIndexView",{},[])
    
    # load secondary chromosomes
    #secondarychroms = getchromosomes( settings['secondarychromdir'] )

    #
    # open alignment file
    #
    if settings['alignmenttype'] == 'axt':
        apv = db.open("AlignmentParView",{'filename':settings['alignmentfile'] % settings['chromosome']},[])

    elif settings['alignmenttype'] == 'maf':
        apv = db.open("MAFXParView",{'filename':settings['alignmentfile'] % settings['chromosome'],
                                     'specieslabels':{settings['basespecies']:0, settings['secondaryspecies']:1},
                                     'base':0,
                                     'type':'maf',
                                     'gapthreshold':100},
                      [chromv])

    elif settings['alignmenttype'] == 'maf3':
        apv = db.open("MAFXParView",{'filename':settings['alignmentfile'] % settings['chromosome'],
                                     'specieslabels':{settings['basespecies']:0,
                                                      settings['secondaryspecies']:1,
                                                      settings['tertiaryspecies']:2},
                                     'base':0,
                                     'type':'maf',
                                     'gapthreshold':100},
                      [chromv])
        
    else:
        print "Unknown alignment type: ", settings['alignmenttype'],"  continuing..."
        apv = Null

    if apv:
        alignmentv = db.open("AlignmentView",{},[apv])
        # get synteny anchors
        if settings['genesynteny']:
            anchors = genesyntenyanchors( apv, segexonv, genev, segtranscriptv )
        else:
            anchors = syntenyanchors( apv, segtrepv, segtev )





#
# Build GC content histogram
#
def gchistogram():

    # make histogram
    bins = int( 1.0 / settings['gchistbinsize'] + 0.5 )
    hist = [0 for i in range( bins )]
    winsize = settings['gcwindowsize']

    # use windows that each half overlap its neighbours
    maxi = len(chromv) - winsize
    pr = Progress()
    for i in range( 0, len(chromv) - winsize, winsize/2 ):

        pr.progress( i, maxi )
        
        sequence = chromv[ i : i+winsize ]
        gcs, nucs = aligntools.cplusg( sequence )
        # allow at most 10% unknown nucleotid
        if nucs >= 0.9 * winsize:
            # calculate the fraction GC, and add noise to remove histogram bin aliasing
            fgc = (gcs + random.random()) / (nucs + 1)
            # calculate bin number; this is always < bins
            bin = int(fgc * bins)
            # add count
            hist[bin] += nucs

    writehistogram( hist, settings['gchistbinsize' ], "GCcontent" )




#
# Calculate the local neutral rates
#
def localrates( generator, chromv):

    window = settings['gcwindowsize']    # window size for local rates
    data = {}

    for igsdata in generator:

        ( igslen,
          gaplen,
          gaplen2,
          actseg,
          seq1,
          chrom2,
          actseg2,
          strand2,
          seq2,
          gapseq,
          gaplen3,
          actseg3,
          chrom3,
          strand3,
          seq3 ) = igsdata
        
        (start1, end1) = actseg
        (start2, end2) = actseg2

        # overlap with AR?
        teoverlap = segtev.overlap( actseg )
        inar = (teoverlap > max(10, igslen-10))

        # overlap with exon?
        inexon = (segexonv.overlap( actseg ) > 0)

        # calculate FDR
        gcclass, gcfraction = getgcclassfrompos( start1 + (igslen//2), chromv )
        bestfdr = calcbestFDR( igslen, gcclass )

        if settings['localratemode'] == '0':
            # use ARs as proxy for neutral sequence
            neutralproxy = inar
        else:
            # use non-exonic non-indel-purified as proxy ('maxfdr' is minimum FDR here)
            neutralproxy = (not inexon) and (bestfdr >= settings['maxfdr'])

        if neutralproxy:

            if seq3 == "":

                ( id, tot, idcg, totcg, idtv, tottv ) = \
                  aligntools.alignidentityCG( seq1, seq2 )

            else:

                (seq1cnt,seq2cnt,seq3cnt,tottv) = \
                                                aligntools.alignidentity3tv( seq1, seq2, seq3 )

                idtv = tottv - (seq1cnt + seq2cnt + seq3cnt)

                (mc0,mc1,mc2,mc012,tot,mccg0,mccg1,mccg2,mccg012,totcg) = \
                                                                        aligntools.alignidentity3( seq1, seq2, seq3 )

                id = tot - (mc0 + mc1 + mc2 + 2*mc012)
                idcg = totcg - (mccg0 + mccg1 + mccg2 + 2*mccg012)

            pos = (start1+end1) // 2
            bin = pos // window
            if not data.has_key(bin):
                data[bin] = [0,0,0,0,0,0]

            data[bin][0] += id     # sequence identity
            data[bin][1] += tot
            data[bin][2] += idcg   # sequence identity without CpG sites
            data[bin][3] += totcg
            data[bin][4] += idtv   # sequence identity w.r.t. transversions
            data[bin][5] += tottv

    entries = data.keys()
    entries.sort()

    file = open(settings['outputfile'],'w')
    file.write("chr\tstart\tend\tid\ttot\tcpgid\tcpgtot\ttvid\ttvtot\n")

    for e in entries:
        file.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % \
                   (settings['chromosome'],\
                    e * window,\
                    (e+1)*window,\
                    data[e][0],\
                    data[e][1],\
                    data[e][2],\
                    data[e][3],\
                    data[e][4],\
                    data[e][5]) )
    file.close()




#
# Generates tuples (len, gap, seg0, seq0, chr1, seg1, strand1, seq1), where
#  len = inter-gap length
#  gap = +/- length of gap preceding this intergap segment (or 0 if start of alignable sequence)
#  chr0, pos0, strand0 = position on base genome
#  chr1, pos1, strand1 = position on other species' genome
#
# This is not a function but a generator.  However it is called like a function and, conceptually
# at least, returns a long list of objects (the successive arguments of 'yield')
#
def pairwise_igs_generator( alignmentparv, alignmentv, chromv, anchors ):

    alignable = 0
    lastpos = lastpos2 = 0
    lastchromstrand2 = ""
    maxgap = settings['maxgap']

    pr = Progress()
    for idx in range( len(alignmentparv) ):

        pr.progress( idx, len(alignmentparv) )

        # get coordinates of alignment; and alignment itself
        alignpar = alignmentparv[idx]
        alignment = alignmentv[idx]

        # get start and end coordinate on base species
        start1 = alignmentparv[idx][ alignmentparv.fields['start1'] ]
        end1 = alignmentparv[idx][ alignmentparv.fields['end1'] ]
        start2 = alignmentparv[idx][ alignmentparv.fields['start2'] ]
        end2 = alignmentparv[idx][ alignmentparv.fields['end2'] ]
        chrom2 = alignmentparv[idx][ alignmentparv.fields['chr2'] ]
        strand2 = alignmentparv[idx][ alignmentparv.fields['strand2'] ]

        alignable += (end1-start1)

        # get intergap segments (igss), and lists of start positions on the base and second species
        (igss,plist1,plist2,gaplist) = aligntools.ungappedalignments(alignment[0],alignment[1], returnidx = True)

        # Bail out if there's no synteny
        if not issyntenous( start1, start2, chrom2, strand2, anchors ):
            lastchromstrand2 = ""
            continue
        
        # loop over igs
        for igidx in range( len( igss ) ):

            igs = igss[igidx]

            # the segments considered
            actseg  = (plist1[igidx]+start1, plist1[igidx]+start1+len(igs[0]))
            actseg2 = (plist2[igidx]+start2, plist2[igidx]+start2+len(igs[1]))
            igslen = Seg.len( actseg )

            # compute length gap preceding this IGS
            curpos = actseg[0]
            curpos2 = actseg2[0]
            if lastpos > curpos - maxgap and lastpos <= curpos and \
               lastpos2 > curpos2 - maxgap and lastpos2 <= curpos2 and chrom2 + strand2 == lastchromstrand2:
                gaplen = curpos - lastpos
                gaplen2 = curpos2 - lastpos2
            else:
                if chrom2 + strand2 != lastchromstrand2:
                    #print "chromstrand",chrom2,strand2
                    pass
                elif lastpos2 < curpos2 - maxgap:
                    #print "maxgap",curpos2-lastpos2
                    pass
                elif lastpos2 > curpos2:
                    #print "backup",lastpos2-curpos2
                    pass
                gaplen = 0
                gaplen2 = 0
            lastpos = actseg[1]
            lastpos2 = actseg2[1]
            lastchromstrand2 = chrom2 + strand2

            # get gap sequence preceding current gap
            if igidx > 0 and igidx <= len(gaplist):
                gapseq = gaplist[igidx-1]
            else:
                gapseq = ""

            gaplen3 = 0
            actseg3 = (0,0)
            chrom3 = ""
            strand3 = "+"

            # return the data
            yield ( len(igs[0]),
                    gaplen,
                    gaplen2,
                    actseg,
                    igs[0],
                    chrom2,
                    actseg2,
                    strand2,
                    igs[1],
                    gapseq,
                    gaplen3,
                    actseg3,
                    chrom3,
                    strand3,
                    "",
                    False)

    # end with a sentinel, to flush data
    yield( 0, 0, 0, (0,0), "", "", (0,0), "+", "", "", 0, (0,0), "", "+", "", False )


#
# Same, but uses mafxparview
#
def triple_igs_generator( alignmentparv, alignmentv, chromv, anchors ):

    alignable = 0
    lastpos = lastpos2 = lastpos3 = 0
    lastchromstrand2 = lastchromstrand3 = ""
    maxgap = settings['maxgap']

    pr = Progress()
    for idx in range( len(alignmentparv) ):

        pr.progress( idx, len(alignmentparv) )

        # get coordinates of alignment; and alignment itself
        alignpar = alignmentparv[idx]
        alignment = alignmentv[idx]

        # get start and end coordinate on base species
        start1 = alignmentparv[idx][ alignmentparv.fields['start1'] ]
        end1 = alignmentparv[idx][ alignmentparv.fields['end1'] ]

        start2 = alignmentparv[idx][ alignmentparv.fields['start2'] ]
        end2 = alignmentparv[idx][ alignmentparv.fields['end2'] ]
        chrom2 = alignmentparv[idx][ alignmentparv.fields['chr2'] ]
        strand2 = alignmentparv[idx][ alignmentparv.fields['strand2'] ]

        start3 = alignmentparv[idx][ alignmentparv.fields['start3'] ]
        end3 = alignmentparv[idx][ alignmentparv.fields['end3'] ]
        chrom3 = alignmentparv[idx][ alignmentparv.fields['chr3'] ]
        strand3 = alignmentparv[idx][ alignmentparv.fields['strand3'] ]

        alignable += (end1-start1)

        # get intergap segments (igss), and lists of start positions on the base and second species
        (igss,plist1,plist2,plist3,gaplist) = aligntools.ungappedalignments3(alignment[0],alignment[1], alignment[2], returnidx = True)

        # Bail out if there's no synteny (only check synteny between species 1 and 2)
        if not issyntenous( start1, start2, chrom2, strand2, anchors ):
            lastchrom2 = ""
            lastchrom3 = ""
            continue

        # loop over igs
        for igidx in range( len( igss ) ):

            igs = igss[igidx]

            # the segments considered
            actseg  = (plist1[igidx]+start1, plist1[igidx]+start1+len(igs[0]))
            actseg2 = (plist2[igidx]+start2, plist2[igidx]+start2+len(igs[1]))
            actseg3 = (plist3[igidx]+start3, plist3[igidx]+start3+len(igs[2]))
            igslen = Seg.len( actseg )

            # compute length gap preceding this IGS
            curpos = actseg[0]
            curpos2 = actseg2[0]
            curpos3 = actseg3[0]
            if lastpos > curpos - maxgap and lastpos <= curpos and \
               lastpos2 > curpos2 - maxgap and lastpos2 <= curpos2 and chrom2 + strand2 == lastchromstrand2 and \
               lastpos3 > curpos3 - maxgap and lastpos3 <= curpos3 and chrom3 + strand3 == lastchromstrand3:
                gaplen = curpos - lastpos
                gaplen2 = curpos2 - lastpos2
                gaplen3 = curpos3 - lastpos3
            else:
                if chrom2 + strand2 != lastchromstrand2:
                    #print "chromstrand",chrom2,strand2
                    pass
                elif lastpos2 < curpos2 - maxgap:
                    #print "maxgap",curpos2-lastpos2
                    pass
                elif lastpos2 > curpos2:
                    #print "backup",lastpos2-curpos2
                    pass
                gaplen = 0
                gaplen2 = 0
                gaplen3 = 0
            lastpos = actseg[1]
            lastpos2 = actseg2[1]
            lastpos3 = actseg3[1]
            lastchromstrand2 = chrom2 + strand2
            lastchromstrand3 = chrom3 + strand3

            # get gap sequence preceding current gap
            if igidx > 0 and igidx <= len(gaplist):
                gapseq = gaplist[igidx-1]
            else:
                gapseq = ""

            # return the data (only for species 1 and 2)
            yield ( len(igs[0]),  #0
                    gaplen,       #1
                    gaplen2,      #2
                    actseg,       #3
                    igs[0],       #4
                    chrom2,       #5
                    actseg2,      #6
                    strand2,      #7
                    igs[1],       #8
                    gapseq,       #9
                    gaplen3,      #10
                    actseg3,      #11
                    chrom3,       #12
                    strand3,      #13
                    igs[2],       #14
                    False)        #15

    # end with a sentinel, to flush data
    yield( 0, 0, 0, (0,0), "", "", (0,0), "+", "", "", 0, (0,0), "", "+", "", False )



# generates IGS that have (or have no) intersection with segmentv
def partial_igs_generator( igs_generator, segmentv, haveintersection ):

    for igsdata in igs_generator:

        actseg = igsdata[3]
        actseg2= igsdata[6]
        actseg3= igsdata[11]
        if haveintersection:

            overlap = segmentv.overlap( actseg )
            if overlap ==0 or overlap == Seg.len(actseg):
                if (overlap == 0 and haveintersection) or (overlap == Seg.len(actseg) and not haveintersection):
                    continue
                else:
                    yield igsdata
                    continue
            removed = Seg.len(actseg) - overlap
            if (segmentv.overlap( (actseg[0],actseg[0]+1) ) == 1) == haveintersection:
                # right end should be removed; sticky end
                yield ( igsdata[0]-removed,
                        igsdata[1],
                        igsdata[2],
                        ( actseg[0], actseg[1]-removed ),
                        igsdata[4][:-removed],
                        igsdata[5],
                        ( actseg2[0], actseg2[1]-removed ),
                        igsdata[7],
                        igsdata[8][:-removed],
                        igsdata[9],
                        igsdata[10],
                        ( actseg3[0], actseg3[1]-removed ),
                        igsdata[12],
                        igsdata[13],
                        igsdata[14][:-removed],
                        True )
                continue
            else:
                # left end should be removed; non-sticky end
                yield ( igsdata[0]-removed,
                        igsdata[1],
                        igsdata[2],
                        ( actseg[0]+removed, actseg[1] ),
                        igsdata[4][removed:],
                        igsdata[5],
                        ( actseg2[0]+removed, actseg2[1]),
                        igsdata[7],
                        igsdata[8][removed:],
                        igsdata[9],
                        igsdata[10],
                        ( actseg3[0]+removed, actseg3[1] ),
                        igsdata[12],
                        igsdata[13],
                        igsdata[14][removed:],
                        False )
                continue
                


def igs_generator( alignmentparv, alignmentv, chromv, anchors ):

    if alignmentparv.name == "MAFXParView" and len(alignmentparv.labels) == 3:
        gen = triple_igs_generator( alignmentparv, alignmentv, chromv, anchors )
    else:
        gen = pairwise_igs_generator( alignmentparv, alignmentv, chromv, anchors )
    subset = settings['subset']
    if subset == "":
        return gen
    if subset == "+exon":
        return partial_igs_generator( gen, segexonv, True )
    if subset == "-exon":
        return partial_igs_generator( gen, segexonv, False )
    if subset == "+te":
        return partial_igs_generator( gen, segtev, True )
    if subset == "-te":
        return partial_igs_generator( gen, segtev, False )
    print "Error: don't understand 'subset' keyword: ",subset
    sys.exit(1)



#
# Helper routine for igshistogram
#
def update_igshist(data, histograms, avgconserved, avgte, avgexon, longigs, saveigs, bestigs, dumponlysegs, file):
                if data['len'] < settings['igshistsize']:
                    histograms[ data['gc'] ][ data['len'] ] += 1
                    avgconserved[ data['gc'] ][ data['len'] ] += (data['idtv']+random.random()) / (data['tottv']+1.0)
                    avgte[ data['gc'] ][ data['len'] ] += data['te'] / (data['len']+0.01)
                    avgexon[ data['gc'] ][ data['len'] ] += data['exon'] / (data['len']+0.01)
                else:
                    longigs.append( (data['len'],
                                     data['gc'],
                                     settings['chromosome'],
                                     data['start'],
                                     data['idtv'] / (0.01+data['tottv']),
                                     data['te'] / (0.01+data['len']),
                                     data['exon'] / (0.01+data['len']) ) )
                # write IGS if appropriate
                if saveigs:
                    include = True           # set default
                    
                    if ratev:
                        if useTV:
                            neutid = neutralid( ratev, rateidxv, ratetots, data['start'], 'tvid', 'tvtot')
                            neutsd = math.sqrt( neutid * (1.0-neutid) / (data['tottv'] + 0.01) )
                            curid = (data['idtv'] + random.random()) / (data['tottv'] + 1.0)
                        else:
                            neutid = neutralid( ratev, rateidxv, ratetots, data['start'], 'cpgid', 'cpgtot')
                            neutsd = math.sqrt( neutid * (1.0-neutid) / (data['totcg'] + 0.01) )
                            curid = (data['idcg'] + random.random()) / (data['totcg'] + 1.0)
                        if curid > neutid - sdfac * neutsd:   # not fast evolving
                            include = False

                    bestfdr = calcbestFDR( data['len'], data['gc'] )
                    
                    if bestfdr > settings['maxfdr']:
                        include = False
                        
                    if include:
                        if dumponlysegs:
                            file.write("%s\t%s\t%s\n" % (settings['chromosome'],data['start'],data['start']+data['len']))
                        else:
                            bestigs.append((data['start'], data['len'], bestfdr))
                    



#
# Build the inter-gap segment histogram, and compute some statistics
#

def igshistogram( igs_generator, chromv, saveigs = False ):

    counts = {'chromosome':0,    # total length
              'known':0,         # of above, only non-N sites
              'syntenous':0,     # of above, those with conserved synteny w.r.t. second species
              'homologous':0,    # of above, those part of an alignment (excluding gaps)
              'partconserved':0, # of above, those non-transversed
               }

    counts['chromosome'] += len(chromv)

    # the igs histograms:
    histograms = [ [ 0 for i in range(settings['igshistsize']) ] for j in range(len(gcbins)+1) ]

    # igs longer than igshistsize are stored here
    longigs = []

    # average conservation
    avgconserved = [ [ 0 for i in range(settings['igshistsize']) ] for j in range(len(gcbins)+1) ]

    # average TE overlap
    avgte = [ [ 0 for i in range(settings['igshistsize']) ] for j in range(len(gcbins)+1) ]

    # average exon overlap
    avgexon = [ [ 0 for i in range(settings['igshistsize']) ] for j in range(len(gcbins)+1) ]

    # running count of igs-related data: length, idtv, tottv (=len), teoverlap, exonoverlap, gapte, gaptot;
    # and static data (start; gcclass)
    data = {'len':0, 'idtv':0, 'tottv':0, 'idcg':0, 'totcg':0, 'te':0, 'exon':0, 'start':-1, 'gc':0, 'sticky':False}
    stickylist = [ copy.deepcopy(data) for i in range(len(gcbins)+1) ]

    # igs length thresholds
    bestigs = []
    file = False
    if saveigs:
        file = open(settings['outputfile'],'w')

    if ratev:
        dumponlysegs = True
    else:
        dumponlysegs = False

    print "DumpOnlySegs = ",dumponlysegs
    
    for igsdata in igs_generator:

        ( igslen,
          gaplen,
          gaplen2,
          actseg,
          seq1,
          chrom2,
          actseg2,
          strand2,
          seq2,
          gapseq,
          gaplen3,
          actseg3,
          chrom3,
          strand3,
          seq3,
          stickyend) = igsdata

        (start1, end1) = actseg
        (start2, end2) = actseg2

        counts['homologous'] += igslen
        counts['syntenous'] += igslen + gaplen    # 0 means: gap too large, synteny break, or gap in opposite sequence

        # get GC class
        gcclass, gcfraction = getgcclassfrompos( start1 + (igslen//2), chromv )

        # calculate number conserved and number non-transversing sites
        mode = settings['localratemode']
        sdfac = 0                    # standard mode: > neutral rate, two way rate
        useTV = True                 # standard: use transversions
        if mode in "14": sdfac = 1   # mode 1 and 3: > neutral rate + 1 SD
        if mode in "25":
            useTV = False
        if mode in "012":            # two-way rates
            ( id, tot, idcg, totcg, idtv, tottv ) = aligntools.alignidentityCG( seq1, seq2 )
        else:                        # 3,4,5 = three-way rates
            ( c0, c1, c2, tottv ) = aligntools.alignidentity3tv( seq1, seq2, seq3 )
            idtv = tottv - (c0+c1+c2)
            (mc0,mc1,mc2,mc012,tot,mccg0,mccg1,mccg2,mccg012,totcg) = aligntools.alignidentity3( seq1, seq2, seq3 )
            idcg = totcg - (mccg0+mccg1+mccg2+2*mccg012)
        counts['partconserved'] += idtv

        # calculate overlap with TEs and exons
        teoverlap = segtev.overlap( actseg )
        exonoverlap = segexonv.overlap( actseg )

        # setup empty data element is necessary
        if data['start'] == -1:
            for key in  ['len','idtv','idcg','tottv','totcg','te','exon']:
                data[key] = 0
            data['start'] = start1
            data['gc'] = gcclass
            data['sticky'] = False
                
        # accumulate if the gap on the left should be skipped (in which case, any change in GC category is ignored)
        if gaplen in settings['allowedgaps'] or \
           gaplen2 in settings['allowedgaps'] or \
           (gaplen>0 and gaplen < abs(settings['maxallowedgap'])) or \
           (gaplen2>0 and gaplen2 < abs(settings['maxallowedgap'])) or \
           (gaplen>0 and settings['maxallowedgap'] > 0) or \
           (gaplen2>0 and settings['maxallowedgap'] < 0):
            data['len'] += igslen
            data['idtv'] += idtv
            data['idcg'] += idcg
            data['tottv'] += tottv
            data['totcg'] += totcg
            data['te'] += teoverlap
            data['exon'] += exonoverlap
            data['sticky'] = stickyend
            if stickyend:
                stickylist[ data['gc'] ] = copy.deepcopy(data)
                data['start'] = -1                 # clear data
        else:
            # true IGS; prepare to store -previous- IGS in data[]
            # First see if it should be added to an earlier IGS with a sticky end
            stickyigs = stickylist[ data['gc'] ]
            if stickyigs['sticky']:
                for key in ['len','idtv','idcg','tottv','totcg','te','exon']:
                    data[key] += stickyigs[key]
                data['start'] = stickyigs['start']
                # remove previous sticky igs
                stickyigs['sticky'] = False
            # Then see if data[] igs itself has a sticky end
            if data['sticky']:
                stickylist[ data['gc'] ] = copy.deepcopy(data)
                data['start'] = -1                 # clear data
            else:
                # If not sticky, store the histogram counts; also store average conservation and other statistics
                update_igshist(data, histograms, avgconserved, avgte, avgexon, longigs, saveigs, bestigs, dumponlysegs, file)

                # store new igs
                data['len'] = igslen
                data['idtv'] = idtv
                data['idcg'] = idcg
                data['tottv'] = tottv
                data['totcg'] = totcg
                data['te'] = teoverlap
                data['exon'] = exonoverlap
                data['start'] = start1
                data['gc'] = gcclass

    # done - store dangling IGS in data and stickylist
    if data['start'] != -1:
        stickyigs = stickylist[ data['gc'] ]
        if stickyigs['sticky']:
            for key in ['len','idtv','idcg','tottv','totcg','te','exon']:
                data[key] += stickyigs[key]
            data['start'] = stickyigs['start']
            # remove previous sticky igs
            stickyigs['sticky'] = False
        update_igshist(data, histograms, avgconserved, avgte, avgexon, longigs, saveigs, bestigs, dumponlysegs, file)

    for data in stickylist:
        if data['sticky']:
            update_igshist(data, histograms, avgconserved, avgte, avgexon, longigs, saveigs, bestigs, dumponlysegs, file)
                
    # write IPS file
    if saveigs:
        if not dumponlysegs:
            file.write("track name=\"Conservation\" description=\"Indel-based Conservation\" visibility=3 color=0,60,120 useScore=1\n")
            for (igsstart,igslen,igsfdr) in bestigs:
                if igsfdr < 0.011:
                    color = 500   # 1% FDR or better
                else:
                    color = 999   # >1% FDR
                file.write("%s\tIGS\tIGS\t%s\t%s\t%s\t.\t.\t%s\n" % (settings['chromosome'],
                                                                     igsstart,
                                                                     igsstart+igslen-1,
                                                                     color,
                                                                     igsfdr))
        file.close()
        return

    # done - write histograms
    histdata = []
    for i in range(len(histograms)):
        h = histograms[i]
        histdata = writehistogram( h, 1, "igslen", histdata = histdata, countlabel = "count_GC"+str(i) )
    for i in range(len(avgconserved)):
        h = avgconserved[i]
        histdata = writehistogram( h, 1, "igslen", histdata = histdata, countlabel = "totid_GC"+str(i) )
    for i in range(len(avgte)):
        h = avgte[i]
        histdata = writehistogram( h, 1, "igslen", histdata = histdata, countlabel = "totte_GC"+str(i) )
    # analysis code relies upon location of exons in table
    if exoncol != 3:
        print "Exon column at wrong position!"
        sys.exit(1)
    for i in range(len(avgexon)):
        h = avgexon[i]
        histdata = writehistogram( h, 1, "igslen", histdata = histdata, countlabel = "totexon_GC"+str(i) )
    writehistogram( None, 0, histdata = histdata )

    # write summary statistics
    print
    print counts
    print "long igs:"
    print longigs

    totallong = 0
    for igs in longigs:
        totallong += igs[0]
    print "Total size:",totallong





def gaphistogram( igs_generator, chromv ):

    counts = {'chromosome':0,    # total length
              'known':0,         # of above, only non-N sites
              'syntenous':0,     # of above, those with conserved synteny w.r.t. second species
              'homologous':0,    # of above, those part of an alignment (excluding gaps)
              'partconserved':0, # of above, those non-transversed
              'gaptot':0,        # total length of gaps in base chromosome
              'gapte':0}         # of above, those overlapping TEs

    counts['chromosome'] += len(chromv)

    # histogram of gap lengths
    gaplens = [ [ 0 for i in range(-settings['gaphistsize'],settings['gaphistsize']) ] for j in range(len(gcbins)+1) ]

    # TE overlap with gaps
    # Keep track of how often repeats are aligned, and how often they are opposite a gap.
    gapkeys = settings['gaprepeats']
    if gapkeys == "":
        gapkeys = ["GC%i" % i for i in range(len(gcbins)+1)]
    else:
        gapkeymap = {}
        repeataligncounts = {}
        repeatgapcounts = {}
        for i in range(len(gapkeys)):
            gapkeymap[ gapkeys[i] ] = i
            repeataligncounts[ gapkeys[i] ] = [1,1]          # counter for aligning and gapped repeats
    gaptes = [ [ 0 for i in range(-settings['gaphistsize'],settings['gaphistsize']) ] for j in gapkeys ]

    for igsdata in igs_generator:

        ( igslen,
          gaplen,
          gaplen2,
          actseg,
          seq1,
          chrom2,
          actseg2,
          strand2,
          seq2,
          gapseq,
          gaplen3,
          actseg3,
          chrom3,
          strand3,
          seq3) = igsdata
        
        (start1, end1) = actseg
        (start2, end2) = actseg2

        counts['homologous'] += igslen
        counts['syntenous'] += igslen + gaplen    # 0 means: gap too large, synteny break, or gap in opposite sequence

        # get a section of the base species' chromosome centered on the current IGS
        gcclass, gcfraction = getgcclassfrompos( start1 + (igslen//2), chromv )

        # calculate overlap with TEs
        teoverlap = segtev.overlap( actseg )

        # dump AR sequence if within bounds
        if settings['dumpARs']:
            margin,gclo,gchi = settings['dumpARs']
            if gcfraction>gclo and gcfraction<gchi and Seg.len( actseg ) > 2*margin and teoverlap > Seg.len(actseg)*0.95:
                print "#SEQ1",seq1
                print "#SEQ2",seq2
                
        # calculate overlap of gap (in base species) with tes
        if gaplen > 0:
            gapte = segtev.overlap( (actseg[0]-gaplen, actseg[0]) )
        else:
            gapte = 0
        counts['gapte'] += gapte
        counts['gaptot'] += gaplen

        # accumulate gap length histogram
        if gaplen > 0:
            gapidx = gaplen + settings['gaphistsize']
            if gapidx < 2*settings['gaphistsize']:
                gaplens[ gcclass ][ gapidx ] += 1

        if gaplen2 > 0:
            gapidx = -gaplen2 + settings['gaphistsize']
            if gapidx >= 0:
                gaplens[ gcclass ][ gapidx ] += 1

        # deal with gc-specific or repeat-specific gap counts
        if gaplen>0:
            if settings['gaprepeats'] == "":
                gapdata = [ (gcclass, gapte) ]
            else:
                rolist = repeatoverlap( (actseg[0]-gaplen, actseg[0]), segtev, tev )
                gapdata = []
                for (overlap, repcls, repnam) in rolist:
                    if repnam in settings['gaprepeats']:
                        gapdata.append( ( gapkeymap[repnam], overlap) )
                        # add to the repeats-to-gap counts
                        repeataligncounts[ repnam ][0] += overlap
                # also accumulate counts for repeat-to-sequence counts
                rolist = repeatoverlap( actseg, segtev, tev )
                for (overlap, repcls, repnam) in rolist:
                    if repnam in settings['gaprepeats']:
                        repeataligncounts[ repnam ][1] += overlap
            for (repkey,overlap) in gapdata:
                gapidx = gaplen + settings['gaphistsize']
                if gapidx < 2*settings['gaphistsize']:
                    gaptes[ repkey ][ gapidx ] += (overlap + 0.0)/abs(gaplen)
                
    # done - write gap length histogram
    histdata = []
    for i in range(len(gaplens)):
        h = gaplens[i]
        histdata = writehistogram( h, 1, "gaplen", histdata = histdata, countlabel = "gapcount_GC"+str(i) )
    for i in range(len(gaptes)):
        h = gaptes[i]
        histdata = writehistogram( h, 1, "gaplen", histdata = histdata, countlabel = gapkeys[i] )
    writehistogram( None, 0, histdata = histdata )

    writedicthistogram( repeataligncounts, 'gaprepoutfile' )

    # write summary statistics
    print
    print counts
    print
    for key in settings['gaprepeats']:
        if key != "":
            print "%12s %12s %12s %12s" % (key, \
                                           repeataligncounts[key][0], \
                                           repeataligncounts[key][1], \
                                           repeataligncounts[key][0]/(repeataligncounts[key][0]+repeataligncounts[key][1]) )
    



def simulate_nuc( nuc, a, b, time ):

    nuc = nuc.upper()
    if nuc == 'A':
        nuc = 0
    elif nuc == 'G':
        nuc = 1
    elif nuc == 'C':
        nuc = 2
    elif nuc == 'T':
        nuc = 3
    else:
        # N
        nuc = 0

    t = 0
    
    while t < time:
 
        rate = a + 2*b
        waitingtime = -math.log(random.random()) / rate
        t += waitingtime
        
        if t < time:
            if random.random() * rate < a:
                # transition
                nuc = (nuc & ~1) + (1 - (nuc&1))
            else:
                # transversion
                nuc = (nuc | 1)
                nuc = 3-nuc + random.randrange(2)

    return "AGCT"[nuc]
 



def find_locus( s1, ancestral, length ):

    # first build current sequence
    seq = []
    for i in range(len(s1)):
        if s1[i] >= 0: seq.append( ancestral[s1[i] % len(ancestral)].upper() )

    # be sure
    if length >= len(seq):
        length = random.randrange(len(seq))

    # build self-similarity array over distance 'length'
    similarity = [0 for i in range(len(seq))]
    for i in range(len(seq)):
        sim = 1
        if seq[i] == seq[i - length]:
            sim = 4
        else:
            if seq[i] in "AG":
                if seq[i - length] in "AG":
                    sim = 2
            elif seq[i - length] in "CT":
                sim = 2
        for j in range(length):
            similarity[i-j] += sim

    # calculate total
    total = 0
    for i in range(len(seq)):
        total += similarity[i]

    # find location
    rand = random.randrange(total)
    loc = -1
    while rand>=0:
        loc += 1
        rand -= similarity[loc]
    return loc




def simulate_seq( ancestral, gapdb, gaplens, indelrate, substitutionrate_a, substitutionrate_b, flank, gcclass ):

    INSERTRANDOM = True

    # make sure we have a reasonable amount of sequence to draw from
    if len(gapdb) < 10*flank:
        return (None,None,None,None)

    seqlen = len(ancestral)
    s1 = [i for i in range(seqlen)]             # build a gapless alignment (-1 is a gap)
    s2 = [i for i in range(seqlen)]

    indels = 0
    gaps = 0
    time = 0
    while time<1.0:
 
        # calculate the number of nucleotides in sequence 1
        tot = 0
        for n in s1:
            if n>=0:  tot += 1
 
        # calculate the indel rate for sequence 1
        rate = tot * indelrate
 
        # get time of next-event
        time += -math.log(random.random())/rate
        if time < 1.0:
 
            # find length of gap
            totgaps = 0
            for g in gaplens:
                totgaps += g
            r = random.randrange(totgaps)
            length = -1
            while r >= 0:
                length += 1
                r -= gaplens[length]

            if INSERTRANDOM:
                # find (random) location of gap; point to first site
                loc = int(tot * random.random())
            else:
                # find location with a preference for repetitive regions
                loc = find_locus( s1, ancestral, length )

            idx = 0
            while loc>=0 and idx<len(s1):
                if s1[idx] >= 0:  loc -= 1
                idx += 1
            idx -= 1
                
            if random.random() < 0.5:                       # insertion
                start = s1[idx]
                insert = range(start,start+length)          # build piece to insert
                gap = [-2 for i in range(length)]           # build gap (signal insertion)
                s1 = s1[:idx] + insert + s1[idx:]           # build new sequence
                s2 = s2[:idx] + gap    + s2[idx:]
            else:                                           # deletion
                while (length>0) and (idx < len(s1)):
                    if s1[idx] >= 0:
                        s1[idx] = -1
                        length -= 1
                    idx += 1

            indels += 1

    # now thread gaps, ancestral sequence, and descendant sequence, through the result
    gapidx = random.randrange( len(gapdb ) )
    prevs1 = 0
    prevs2 = 0
    for idx in range(len(s1)):

        if s1[idx]>=0 or s2[idx]>=0:

            if (s1[idx]<0 or s2[idx]<0) and (prevs1 >= 0 and prevs2 >= 0):
                gaps += 1
                gapidx = random.randrange( len( gapdb ) )

            prevs1 = s1[idx]
            prevs2 = s2[idx]

            if s2[idx] == -1:   # deletion of descendant sequence
                s1[idx] = ancestral[s1[idx]]
                s2[idx ] = '-'
            elif s2[idx] <= -2: # insertion in ancestral sequence
                if INSERTRANDOM:
                    s1[idx] = gapdb[gapidx]
                else:
                    fromidx = s1[idx] % len(ancestral)
                    s1[idx] = ancestral[fromidx]
                s2[idx] = '-'
                gapidx = (gapidx + 1) % len(gapdb)
            elif s1[idx] == -1:  # deletion of ancestral sequence
                s1[idx] = '-'
                s2[idx] = ancestral[s2[idx]]
            elif s1[idx] <= -2:  # insertion in descendant sequence
                if INSERTRANDOM:
                    s2[idx] = gapdb[gapidx]
                else:
                    fromidx = s2[idx] % len(ancestral)
                    s2[idx] = ancestral[fromidx]
                s1[idx] = '-'
                gapidx = (gapidx + 1) % len(gapdb)
            else:                # survival
                anc = ancestral[s1[idx]]
                s1[idx] = simulate_nuc( anc, substitutionrate_a, substitutionrate_b, 0.5 )
                s2[idx] = simulate_nuc( anc, substitutionrate_a, substitutionrate_b, 0.5 )

        else:
            s1[idx] = ''
            s2[idx] = ''

    p = random.randrange( len(gapdb) - 4*flank )
    gap = '-' * flank
    f1 = gapdb[ p:p+flank ]+gap
    f2 = gap+gapdb[ p+flank:p+2*flank ]
    f3 = gapdb[ p+2*flank:p+3*flank ]+gap
    f4 = gap+gapdb[ p+3*flank:p+4*flank ]

    return ( f1+(''.join(s1))+f3, f2+(''.join(s2))+f4, indels, gaps )




def simulate( igs_generator, chromv ):

    # gaphistsize - determines largest gap considered non-TE

    amount = 40000000.0
    density = amount / len(chromv)
    
    #indelrate =[0.05188, 0.04798, 0.04668, 0.04553, 0.04548, 0.04551, 0.04542, 0.04534, 0.04505, 0.04560, 0.04610, 0.04685, 0.04808, 0.04849, 0.04950, 0.05092, 0.05303, 0.05548, 0.05893, 0.06474]
    #for i in range(20):
    #    # this accounts for both indel rate -> gap density on the primary
    #    # sequence, and gap density / nuc to gap density / aligned nuc
    #    indelrate[i] *= 1.04

    #indelrate = [0.057243769836515586, 0.051738773792156259, 0.050054640496086128, 0.048675298660123906, 0.048527713321970964, 0.048472291694715362, 0.04836690110543175, 0.048299987977080946, 0.048001192587688882, 0.048616777721188296, 0.049243480846279934, 0.050109784460332339, 0.051516448371689977, 0.052128432882649262, 0.053286174843129648, 0.054944258686371407, 0.057467928891203696, 0.060337418887682619, 0.064312592175012312, 0.071332848947716285]

    #indelrate = [0.05829390326854799, 0.052845878139832173, 0.051255654019652458, 0.050491657289875404, 0.04968030654795115, 0.049632790772854034, 0.049729558918688889, 0.049799350490433346, 0.048918221113782349, 0.049792071917660309, 0.05026930644729681, 0.051450946524277168, 0.052567663055107475, 0.053246423555374733, 0.054596839759249699, 0.056216298299667705, 0.058938869787687802, 0.061811388427936667, 0.06616082634003001, 0.073499528209665377]

    indelrate = [0.0505 for i in range(20)]

    #indelrate = [0.054546037, 0.051700488, 0.05038209, 0.050447592, 0.050837304, 0.050220893, 0.050333182, 0.050224783, 0.050698256, 0.050456973, 0.050570719, 0.050580173, 0.051948489, 0.051640392, 0.052675651, 0.05391877, 0.055880367, 0.058362617, 0.061996064, 0.067028603]
                 
    #substitutionrate_a = 0.19
    #substitutionrate_b = 0.09
    #substitutionrate_a = 0.225
    #substitutionrate_b = 0.105

    substitutionrate_a = 0.218
    substitutionrate_b = 0.109
    sequence_length = 1000
    flanks = 100

    gaphistsize = settings['gaphistsize']
    gapdblen = 10000

    totindel = [0 for i in range(len(gcbins)+1)]
    totgaps = [0 for i in range(len(gcbins)+1)]

    # histogram of gap lengths
    gaplens = [ [ 0 for i in range(gaphistsize) ] for j in range(len(gcbins)+1) ]

    # gapped sequence
    gapdb = [ "" for j in range(len(gcbins)+1) ]

    # "ancestral" sequence
    ancestral = [ "" for j in range(len(gcbins)+1) ]

    output = open(settings['outputfile'],'w')

    for igsdata in igs_generator:

        ( igslen,
          gaplen,
          gaplen2,
          actseg,
          seq1,
          chrom2,
          actseg2,
          strand2,
          seq2,
          gapseq,
          gaplen3,
          actseg3,
          chrom3,
          strand3,
          seq3) = igsdata
        
        (start1, end1) = actseg
        (start2, end2) = actseg2

        exonoverlap = segexonv.overlap( actseg )

        # get a section of the base species' chromosome centered on the current IGS, and calculate GC class
        gcclass, gcfraction = getgcclassfrompos( start1 + (igslen//2), chromv )

        # accumulate gap length histogram
        if gaplen>0 or gaplen2>0:
            if max(gaplen,gaplen2) < gaphistsize:
                gaplens[ gcclass ][ max(gaplen,gaplen2) ] += 1

        # build gap sequence database
        gapdb[ gcclass ] += gapseq
        if len(gapdb[ gcclass ]) > gapdblen and random.random()<0.01:
            gapdb[ gcclass ] = gapdb[ gcclass ][-gapdblen:]

        # build 'ancestral' sequence
        if exonoverlap == 0:
            
            ancestral[ gcclass ] += seq1
            if gaplen < gaphistsize and gaplen2 < gaphistsize and random.random() > 0.5:
                ancestral[ gcclass ] += gapseq

            if len( ancestral[ gcclass ] ) >= sequence_length:

                if random.random() < density and len( ancestral[ gcclass ] ) < sequence_length * 1.1:

                    (a1, a2, indels, gaps) = simulate_seq( ancestral[gcclass],\
                                                           gapdb[gcclass],\
                                                           gaplens[gcclass],\
                                                           indelrate[gcclass],\
                                                           substitutionrate_a,\
                                                           substitutionrate_b,\
                                                           flanks,\
                                                           gcclass )

                    if a1:
                        output.write(">"+str(gcclass)+"\n"+a1+"\n"+a2+"\n\n")

                        totindel[gcclass] += indels
                        totgaps[gcclass] += gaps

                ancestral[ gcclass ] = ""

    print "Totindel ",totindel
    print "Totgaps  ",totgaps
                              





#
# Combine histograms for single chromosomes into a single big histogram
#
def combinehistograms():

    header = ""
    labels = []
    hist = {}

    for filename in settings['--']:

        file = open(filename,'r')
        lines = file.readlines()

        # store or check header line
        if not header:
            header = lines[0]
        else:
            if header != lines[0]:
                print "Warning: Header lines don't match (file=",filename,")"

        # loop over the data lines
        labels = []
        for line in lines[1:]:

            # convert the text line into integers or floats
            line = line[:-1].split('\t')
            line = [line[0]] + map( lambda x: (x.find('.')==-1 and [int(x)] or [float(x)])[0], line[1:] )

            # store label
            labels.append( line[0] )

            # set all counts to 0 if this is the first line
            if line[0] not in hist:
                hist[line[0]] = [0 for i in range(len(line)-1)]

            # accumulate counts
            for i in range(0,len(line)-1):
                hist[line[0]][ i ] += line[ i+1 ]
            
    
    # done - write histograms
    file = open(settings['outputfile'],'a')
    file.write( header )
    for label in labels:
        file.write("%s" % label)
        for count in hist[label]:
            file.write("\t%s" % count)
        file.write('\n')
            
        
        
#
# analyze histograms
#
def analyze():

    histfile = settings['--'][0]
    if histfile == "":
        print "No histogram file provided; use '-- filename' after all the options"
        sys.exit(1)
    numgc = len( gcbins ) + 1
    numgc2= numgc                              # for debugging

    beginstart=10                              # limits for beginning of regression interval
    beginstop=25
    endstart=40                                # limits for end of regression interval
    endstop=100
    if debug:
        beginstart=20
        beginstop=21
        endstart=70                           
        endstop=71
    if settings['regressionstart']:
        startrows = "c(" + settings['regressionstart'] + ")"
        endrows = "c(" + settings['regressionend'] + ")"
        fixedinterval = "TRUE"
    else:
        startrows = str(beginstart) + ":" + str(beginstop)
        endrows = str(endstart) + ":" + str(endstop)
        fixedinterval = "FALSE"

    rpy.r.source( settings["Rfile"] )
    hist = rpy.r("hist <- read.table(\"%s\", header=TRUE)" % histfile)
    an = rpy.r("an <- analyze( hist, 2:%s, exoncol=3*%s, startrows=%s, endrows=%s, plotfile='%s', fixedinterval=%s )" % \
               (numgc2+1,numgc,startrows,endrows,settings['outputfile'],fixedinterval))

    print "Total indel-purified, lower bound:\t", an["totlower"]
    print "Total indel-purified, upper bound:\t", an["totupper"]
    print "Total indel-purified + overhang:\t", an["totips"]
    print "Total aligning:\t", an["totalign"]
    print "Total exonic:\t", an["totexon"]
    print "Gap probability:\t", an["gapprobability"]
    print "Proportion indel-purified, lower bound:\t", an["proplower"]
    print "Proportion indel-purified, upper bound:\t", an["propupper"]
    print "Proportion indel-purified + overhang:\t", an["propips"]
    print "IPS lengths used for NIM fit, lower limits:\t", an["lowlimits"]
    print "IPS lengths used for NIM fit, upper limits:\t", an["uplimits"]
    print "intercepts of NIM fits:\t", an["a"]
    print "slope parameters of NIM fits:\t", an["b"]

    print "Identification of IPSs:"
    print "TotCons'd:\tIPS:     \tUpper:   \tLower:   \tExonic:  \tFDR:\tFDR:\tIGS length cutoff:"

    fdrmap = {}
    fdrrange = range(1,51)
    if debug:
        fdrrange = range(1,51,10)

    for fdr in fdrrange:
        data = rpy.r("fdr1 <- fdrcutoffs( hist, an, %s, exoncol=3*%s)" % ( fdr*0.01, numgc ))
        print "%9.f\t%9.f\t%9.f\t%9.f\t%9.f\t%1.3f\t%1.3f\t%s" % \
              (data['totcons'],
               data['totips'],
               data['totsel_upper'],
               data['totsel_lower'],
               data['totexon'],
               data['fdr'],
               fdr*0.01,
               map(int,data['cutoffs']))
        fdrmap[fdr*0.01] = map(int,data['cutoffs'])

    cutoffs = open(settings['thresholdfile'],'w')
    cutoffs.write( str(fdrmap) )
    cutoffs.close()







##############################################################################################################################
#
#
#  Top-level code
#
#
##############################################################################################################################        



# read command line settings
options = [('-action=', getstring, "{[1] gchistogram, [2] ipshistogram, [3] combinehistograms, [4] analyze, [5] gaphistogram, [6] getips, [7] simulate, [8] localrates}", "none"),
           ('-verbose=', getint, "0, 1 or 2", 2),
           ('-basespecies=', getstring, "[1,2] e.g. hg17 (how species are named in alignment)", "hg17"),
           ('-basechromdir=', getstring, "[1,2] directory of base species chromosomes; must end in /", \
            "/net/cpp-group/gerton/Data/ucsc/hg17/chromosomes/"),
           ('-secondaryspecies=', getstring, "[1,2] e.g. mm5", "mm5"),
           ('-secondarychromdir=', getstring, "[2] directory of secondary species chromosomes",\
            "/net/cpp-group/gerton/Data/ucsc/mm5/chromosomes/"),
           ('-tertiaryspecies=', getstring, "[1,2] e.g. canFam2; use with alignmenttype=maf3", ""),
           ('-chromosome=', getstring, "[1,2] chromosome to process", "chr21"),
           ('-chromlabelprefix=', getstring, "[1,2] prefix for chromosome labels in repeatmasker and gene files",""),
           ('-gcwindowsize=', getint, "[1,2,8] window size for gc content; or window size for local rates", 250),
           ('-gchistbinsize=', getfloat, "[1] histogram bin size for gc content", 0.005),
           ('-outputfile=', getstring, "[1,2,5,8] where output gets sent", "ips.out"),
           ('-thresholdfile=', getstring, "[4,6] in/out file where IGS length cutoffs are sent", "thresholds.txt"),
           ('-gaprepoutfile=', getstring, "[5] where the aligning- and gapped-counts for repeat families get written", "gapreps.out"),
           ('-tandemrepeatfile=', getstring, "[2] columns: ?, chromosome, start, end, period, pid, pindel, .....", ""),
           ('-repeatmaskfile=', getstring, "[2] repeatmasker file", "/net/cpp-group/gerton/Data/ucsc/hg17/chromosomes/%s.fa.out"),
           ('-localratefile=', getstring, "[2,6] local neutral rates", ""),
           ('-localratemode=', getstring, "[6,8] [6]: 0: >neutral transv; 1: >neutral+1SD transv; 2: CpG ID; 3,4,5: triple-alignment local rates; [8]: 0=ARs, 1=heuristic neutral sequence", "0"),
           ('-genefile=', getstring, "[2] gene definitions", "/net/cpp-group/gerton/Data/ucsc/hg17/database/ensGene.txt"),
           ('-subset=', getstring, "[2] genomic subset to analyze: +exon, -exon, +te, -te; default all", ""),
           ('-minsynteny=', getint, "[2] minimal stretch of non-repetitive alignment considered synteny anchor", 250 ),
           ('-maxsyntenygap=', getint, "[2] alignment blocks further apart than this are considered separate anchors", 500),
           ('-genesynteny=', getint, "[2,6] if 1, genes are used as synteny anchors, rather than any unique sequence",0),
           ('-synwindow=', getint, "[2] window considered in conserved synteny with synteny anchor (>= 4000)", 100000 ),
           ('-alignmentfile=', getstring, "[2] directory with .maf or .axt alignment files, from base to secondary",\
             "/net/cpp-group/gerton/Data/ucsc/hg17/vsMm5/axtNet/%s.axt"),
           ('-alignmenttype=', getstring, "[2] axt, maf, or maf3 (=3 species from maf alignment)", "axt"),
           ('-gcbins=', getstring, "[2,4] defines number [2,4] and thresholds [2] of GC content bins", \
             "[0.28,0.304,0.32,0.332,0.348,0.36,0.368,0.38,0.392,0.404,0.416,0.428,0.44,0.456,0.472,0.488,0.508,0.536,0.572]"),
           ('-regressionstart=', getstring, "[4] comma-sep'd list of starts of regr. intervals by GC; default=auto-optimize", ""), \
           ('-regressionend=', getstring, "[4] same for ends of regression intervals", ""),
           ('-igshistsize=', getint, "[2] size of intergap segment histogram", 2000),
           ('-gaphistsize=', getint, "[2,7] size of gap length histogram (2), or max gap considered non-TE (7)", 5000),
           ('-cachedir=', getstring, "[1,2] directory for database cache", "/net/cpp-group/gerton/Localdata/DBView/"),
           ('-maxgap=', getint, "[2] maximum gap still considered an indel rather than synteny break", 10000),
           ('-allowedgaps=', getstring, "[2] comma-separated list of gap lengths that will be ignored", ""),
           ('-maxallowedgap=', getint, "[2] maximum gap length that will be ignored; pos/neg=ignore insertions in base/other", 0),
           ('-maxfdr=', getfloat, "[6] maximum false discovery rate in IPS file", 0.1001),
           ('-Rfile=', getstring, "[4] location of the analyze.R file", "analyze.R"),
           ('-gaprepeats=', getstring, "[2] comma-separated repeat families to keep counts for; if empty, GC classes are used", ""),
           ('-dumpARs=', getstring, "[5] 'margin,gclo,gchi'; margin=number of nucs removed, gclo/hi GC bounds", ""),
           ]

settings = {}

args = getopt(sys.argv[1:], options, settings, parsepart = True)
if args == False:
    sys.exit(1)
settings['--'] = args
settings['allowedgaps'] = eval( "[" + settings['allowedgaps'] + "]" )
settings['gaprepeats'] = settings['gaprepeats'].split(',')
settings['dumpARs'] = eval( "[" + settings['dumpARs'] + "]" )


# open database
print "db dir=", settings['cachedir']
db = dbview.dbview( settings['cachedir'] + "/root" )           # create or open the cache database
db.setRootDict( 'cacheDir', settings['cachedir'] )             # set the cache file directory
db.setRootDict( 'verbose', settings['verbose'] )               # print some progress information


# dump command line options
dumpsettings()
gcbins = eval( settings['gcbins'] )


# make outputfile empty
open(settings['outputfile'],'w').close()




#
# do the work
#
print settings['action']
if settings['action'] == 'gchistogram':

    loaddata()
    dopsyco()
    gchistogram()
    
elif settings['action'] == 'igshistogram':

    loaddata()
    dopsyco()
    generator = igs_generator( apv, alignmentv, chromv, anchors )
    igshistogram( generator, chromv )

elif settings['action'] == 'combinehistograms':

    combinehistograms()

elif settings['action'] == 'analyze':

    dopsyco()
    analyze()
    
elif settings['action'] == 'gaphistogram':

    loaddata()
    dopsyco()
    generator = igs_generator( apv, alignmentv, chromv, anchors )
    gaphistogram( generator, chromv )

elif settings['action'] == 'getips':

    loaddata()
    dopsyco()
    generator = igs_generator( apv, alignmentv, chromv, anchors )
    igshistogram( generator, chromv, saveigs = True )
    

elif settings['action'] == 'simulate':

    loaddata()
    dopsyco()
    generator = igs_generator( apv, alignmentv, chromv, anchors )
    simulate( generator, chromv )

elif settings['action'] == 'localrates':

    loaddata()
    dopsyco()
    generator = igs_generator( apv, alignmentv, chromv, anchors )
    localrates( generator, chromv )
    

else:

    print "Nothing to do.."


#
# exit
#
if settings['verbose'] >= 2:
    print
